from __future__ import annotations

from importlib import import_module
from pathlib import Path
from typing import Annotated

try:
    import anyio
except ImportError:
    anyio = None
from docnote import ClcNote

from templatey._types import TemplateParamsInstance
from templatey.environments import AsyncTemplateLoader
from templatey.environments import SyncTemplateLoader


class DictTemplateLoader[L: object](
        AsyncTemplateLoader[L], SyncTemplateLoader[L]):
    """A barebones template loader that simply loads templates from a
    dictionary based on whatever key you supply.
    """
    lookup: Annotated[dict[L, str],
        ClcNote('''
            Provides direct access to the template lookup. Store literal
            template text here using whatever key matches the resource locator
            used on your template definitions.
            ''')]

    def __init__(self, templates: dict[L, str] | None = None):
        if templates is None:
            templates = {}

        self.lookup = templates

    def load_sync(
            self,
            template: type[TemplateParamsInstance],
            template_resource_locator: L
            ) -> str:
        return self.lookup[template_resource_locator]

    async def load_async(
            self,
            template: type[TemplateParamsInstance],
            template_resource_locator: L
            ) -> str:
        return self.lookup[template_resource_locator]


class CompanionFileLoader(AsyncTemplateLoader[str], SyncTemplateLoader[str]):
    """This loader uses string resource locators that correspond to
    files located in the same directory as the underlying template
    definition. For example:

    > my_templates.py
    __embed__: 'code/python'
        from templatey import template
        from templatey.prebaked.template_configs import html

        @template(html, 'custom_template.html')
        class CustomTemplate:
            ...

    That would search for a file called ``custom_template.html`` in the
    same parent directory as ``my_templates.py``.

    This supports both sync and async loading, but async loading
    requires ``anyio`` to be installed (also available via extras,
    ``pip install templatey[async_prebaked]``).
    """

    def load_sync(
            self,
            template: type[TemplateParamsInstance],
            template_resource_locator: str
            ) -> str:
        template_module_name = template.__module__
        template_module = import_module(template_module_name)
        module_file = template_module.__file__

        if module_file is None:
            raise ValueError(
                'CompanionFileLoader can only load templates defined in '
                + 'actual modules!')

        module_dir = Path(module_file).parent
        return (module_dir / template_resource_locator).read_text(
            encoding='utf-8')

    if anyio is not None:
        async def load_async(
                self,
                template: type[TemplateParamsInstance],
                template_resource_locator: str
                ) -> str:
            template_module_name = template.__module__
            template_module = import_module(template_module_name)
            module_file = template_module.__file__

            if module_file is None:
                raise ValueError(
                    'CompanionFileLoader can only load templates defined in '
                    + 'actual modules!')

            # Pyright isn't correctly applying the anyio is not None, hence the
            # ignore
            module_dir = anyio.Path(module_file).parent  # type: ignore
            return await (module_dir / template_resource_locator).read_text(
                encoding='utf-8')



from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass
from dataclasses import field
from types import UnionType
from weakref import ref

from templatey._forwardrefs import PENDING_FORWARD_REFS
from templatey._forwardrefs import ForwardReferenceProxyClass
from templatey._forwardrefs import ForwardRefLookupKey
from templatey._forwardrefs import is_forward_reference_proxy
from templatey._provenance import Provenance
from templatey._slot_tree import ConcreteSlotTreeNode
from templatey._slot_tree import DynamicClassSlotTreeNode
from templatey._slot_tree import PendingSlotTreeContainer
from templatey._slot_tree import PendingSlotTreeNode
from templatey._slot_tree import SlotTreeNode
from templatey._slot_tree import SlotTreeRoute
from templatey._slot_tree import gather_dynamic_class_slots
from templatey._slot_tree import merge_into_slot_tree
from templatey._slot_tree import update_encloser_with_trees_from_slot
from templatey._types import TemplateClass
from templatey._types import TemplateInstanceID
from templatey._types import TemplateParamsInstance

type GroupedTemplateInvocations = dict[TemplateClass, list[Provenance]]
type TemplateLookupByID = dict[TemplateInstanceID, TemplateParamsInstance]
type _SlotAnnotation = (
    TemplateClass
    | UnionType
    | type[ForwardReferenceProxyClass])


@dataclass(slots=True)
class TemplateSignature:
    """This class stores the processed interface based on the params.
    It gets used to compare with the TemplateParse to make sure that
    the two can be used together.

    Not meant to be created directly; instead, you should use the
    TemplateSignature.new() convenience method.

    TODO: we need to get way more consistent about public/private attr
    conventions; this is a holdover from when this was within a public
    module.
    """
    # It's nice to have this available, especially when resolving forward refs,
    # but unlike eg the slot tree, it's trivially easy for us to avoid GC
    # loops within the signature
    template_cls_ref: ref[TemplateClass]
    _forward_ref_lookup_key: ForwardRefLookupKey

    data_names: frozenset[str]
    var_names: frozenset[str]
    content_names: frozenset[str]
    # Note that these are all ONLY the direct nesteds; these do not include
    # anything from deeper in the slot tree.
    slot_names: frozenset[str]
    dynamic_class_slot_names: frozenset[str]

    _dynamic_class_slot_tree: DynamicClassSlotTreeNode
    _ordered_dynamic_class_slot_names: list[str] = field(init=False)
    # Note that these contain all included types, not just the ones on the
    # outermost layer that are associated with the signature. In other words,
    # they include the flattened recursion of all included slots, all the way
    # down the tree
    _slot_tree_lookup: dict[TemplateClass, ConcreteSlotTreeNode]
    _pending_ref_lookup: dict[ForwardRefLookupKey, PendingSlotTreeContainer]

    # I really don't like that we need to remember to recalculate this every
    # time we update the slot tree lookup, but for rendering performance
    # reasons we want this to be precalculated before every call to render.
    included_template_classes: frozenset[TemplateClass] = field(init=False)

    def __post_init__(self):
        self.refresh_included_template_classes_snapshot()
        self.refresh_pending_forward_ref_registration()
        self._ordered_dynamic_class_slot_names = sorted(
            self.dynamic_class_slot_names)

    def refresh_included_template_classes_snapshot(self):
        """Call this when resolving forward references to apply any
        changes made to the slot tree to the template classes snapshot
        we use for increased render performance.
        """
        template_cls = self.template_cls_ref()
        if template_cls is None:
            raise RuntimeError(
                'Template class was garbage collected before template '
                + 'signature, and then signature asked to refresh included '
                + 'classes snapshot?!')

        self.included_template_classes = frozenset(
            {template_cls, *self._slot_tree_lookup})

    def refresh_pending_forward_ref_registration(self):
        """Call this after having resolved forward references (or when
        initially constructing the template signature) to register the
        template class as requiring its forward refs. This is what
        plumbs up the notification code to actually initiate resolving.
        """
        template_cls = self.template_cls_ref()
        if template_cls is None:
            raise RuntimeError(
                'Template class was garbage collected before template '
                + 'signature, and then signature asked to refresh pending '
                + 'forward ref registration?!')

        forward_ref_registry = PENDING_FORWARD_REFS.get()
        for forward_ref in self._pending_ref_lookup:
            forward_ref_registry[forward_ref].add(template_cls)

    @classmethod
    def new(
            cls,
            template_cls: type,
            slots: dict[str, _SlotAnnotation],
            dynamic_class_slot_names: set[str],
            data: dict[str, None],
            vars_: dict[str, type | type[ForwardReferenceProxyClass]],
            content: dict[str, type | type[ForwardReferenceProxyClass]],
            *,
            forward_ref_lookup_key: ForwardRefLookupKey
            ) -> TemplateSignature:
        """Create a new TemplateSignature based on the gathered slots,
        vars, and content. This does all of the convenience calculations
        needed to populate the semi-redundant fields.
        """
        slot_names = frozenset(slots)
        var_names = frozenset(vars_)
        content_names = frozenset(content)

        # Quick refresher: our goal here is to construct a lookup that gets
        # us a route to every instance of a particular template type. In other
        # words, we want to be able to check a template type, and then see all
        # possible getattr() sequences that arrive at an instance of that
        # template type.
        tree_wip: dict[TemplateClass, ConcreteSlotTreeNode]
        tree_wip = defaultdict(SlotTreeNode)
        pending_ref_lookup: \
            dict[ForwardRefLookupKey, PendingSlotTreeContainer] = {}

        concrete_slot_defs, pending_ref_defs = cls._normalize_slot_defs(
            slots.items())

        # Note that order between concrete_slot_defs and pending_ref_defs
        # DOES matter here. Concrete has to come first, because we need to
        # discover all reference loops back to the template class before
        # we can fully define the pending trees.
        for slot_name, slot_type in concrete_slot_defs:
            # In the simple recursion case -- a template defines a slot of its
            # own class -- we can immediately create a reference loop without
            # any pomp nor circumstance.
            # Also: yes, this is resolved at annotation time, and not a forward
            # ref!
            if slot_type is template_cls:
                recursive_slot_tree = tree_wip[slot_type]
                recursive_slot_tree.is_terminus = True
                recursive_slot_tree.is_recursive = True
                recursive_slot_route = SlotTreeRoute.new(
                    slot_name,
                    slot_type,
                    recursive_slot_tree)
                recursive_slot_tree.append(recursive_slot_route)

            # Remember that we expanded the union already, so this is
            # guaranteed to be a single concrete ``slot_type``.
            else:
                offset_tree = PendingSlotTreeNode(
                    insertion_slot_names={slot_name})
                update_encloser_with_trees_from_slot(
                    template_cls,
                    tree_wip,
                    pending_ref_lookup,
                    forward_ref_lookup_key,
                    slot_type,
                    offset_tree,)

        # Note that by "direct" we mean, immediate nested children of the
        # current template_cls, for which we're constructing a signature,
        # and NOT any nested children of nested concrete slots.
        # These plain pending refs are very straightforward, since we don't
        # know anything about them yet (by definition; they're forward refs!).
        for (
            direct_pending_slot_name,
            direct_forward_ref_lookup_key
        ) in pending_ref_defs:
            existing_pending_tree = pending_ref_lookup.get(
                direct_forward_ref_lookup_key)

            if existing_pending_tree is None:
                dest_insertion = PendingSlotTreeNode(
                    insertion_slot_names={direct_pending_slot_name})
                pending_ref_lookup[direct_forward_ref_lookup_key] = (
                    PendingSlotTreeContainer(
                        pending_slot_type=direct_forward_ref_lookup_key,
                        pending_root_node=dest_insertion))
                # Note that we need to include any recursion loops that end
                # up back at the template class, since they would ALSO have
                # the same insertion points. Helpfully, we can just merge in
                # any existing tree for that.
                existing_recursive_self_tree = tree_wip.get(
                    template_cls,
                    SlotTreeNode())
                merge_into_slot_tree(
                    template_cls,
                    None,
                    dest_insertion,
                    existing_recursive_self_tree)

            else:
                (existing_pending_tree
                    .pending_root_node
                    .insertion_slot_names.add(direct_pending_slot_name))

        # Okay, so here's the deal. Every time we add a slot,
        # might create a recursive reference loop (if it had a forward ref to
        # the current enclosing class). So far so good. The problem
        # is, this means that existing slot trees will be incomplete.
        # Therefore, at the end of the process, we always circle back and check
        # for a self-referential slot tree, and if it exists, merge it into
        # all the rest.
        self_referential_recursive_tree = tree_wip.get(template_cls)
        if self_referential_recursive_tree is not None:
            for slot_type, slot_tree in tree_wip.items():
                if slot_type is not template_cls:
                    merge_into_slot_tree(
                        template_cls,
                        slot_type,
                        slot_tree,
                        self_referential_recursive_tree)

        # Oh thank god.
        tree_wip.default_factory = None
        return cls(
            _forward_ref_lookup_key=forward_ref_lookup_key,
            template_cls_ref=ref(template_cls),
            slot_names=slot_names,
            dynamic_class_slot_names=frozenset(dynamic_class_slot_names),
            _dynamic_class_slot_tree=gather_dynamic_class_slots(
                template_cls,
                dynamic_class_slot_names,
                tree_wip),
            data_names=frozenset(data),
            var_names=var_names,
            content_names=content_names,
            _slot_tree_lookup=tree_wip,
            _pending_ref_lookup=pending_ref_lookup)

    @classmethod
    def _normalize_slot_defs(
            cls,
            slot_defs: Iterable[tuple[str, _SlotAnnotation]]
            ) -> tuple[
                list[tuple[str, TemplateClass]],
                list[tuple[str, ForwardRefLookupKey]]]:
        """The annotations we get "straight off the tap" (so to speak)
        of the template class can be:
        ++  unions
        ++  type aliases, though we don't quite support these yet
            (though this function should make it relatively
            straightforward to do so)
        ++  concrete backrefs to existing templates
        ++  pending forward refs to not-yet-defined templates

        This function is responsible for giving us a clean, uniform way
        of representing them. First, it expands all unions etc into
        multiple slot paths. Second, it splits the concrete from the
        pending nodes, returning them separately.
        """
        concrete_slots = []
        pending_refs = []

        flattened_defs: \
            list[tuple[
                str,
                TemplateClass | type[ForwardReferenceProxyClass]]] = []
        for slot_name, slot_annotation in slot_defs:
            # Note that this still might contain a heterogeneous mix of
            # template classes and forward refs! Hence flattening first.
            if isinstance(slot_annotation, UnionType):
                for slot_type in slot_annotation.__args__:
                    flattened_defs.append((slot_name, slot_type))

            else:
                flattened_defs.append((slot_name, slot_annotation))

        # Again, this looks redundant at first glance, but the point was to
        # normalize unions into single types, whether concrete or pending
        for slot_name, flattened_slot_annotation in flattened_defs:
            if is_forward_reference_proxy(flattened_slot_annotation):
                forward_ref = flattened_slot_annotation.REFERENCE_TARGET
                pending_refs.append((slot_name, forward_ref))

            else:
                concrete_slots.append((slot_name, flattened_slot_annotation))

        return concrete_slots, pending_refs

    def resolve_forward_ref(
            self,
            lookup_key: ForwardRefLookupKey,
            resolved_template_cls: TemplateClass
            ) -> None:
        """Notifies a dependent class (one that declared a slot as a
        forward reference) that the reference is now available, thereby
        causing it to resolve the forward ref and remove it from its
        pending trees.
        """
        enclosing_template_cls = self.template_cls_ref()
        if enclosing_template_cls is None:
            raise RuntimeError(
                'Template class was garbage collected before template '
                + 'signature, and then signature asked to resolve forward '
                + 'ref?!')

        update_encloser_with_trees_from_slot(
            enclosing_template_cls,
            self._slot_tree_lookup,
            self._pending_ref_lookup,
            self._forward_ref_lookup_key,
            resolved_template_cls,
            self._pending_ref_lookup.pop(lookup_key).pending_root_node,)

        self.refresh_included_template_classes_snapshot()
        self.refresh_pending_forward_ref_registration()

        # Okay, so here's the deal. Every time we resolve a forward ref, we
        # might create a recursive reference loop. So far so good. The problem
        # is, this means that existing slot trees will be incomplete.
        # therefore, at the end of the process, we always circle back and check
        # for a self-referential slot tree, and if it exists, merge it into
        # all the rest.
        self_referential_recursive_tree = self._slot_tree_lookup.get(
            enclosing_template_cls)
        if self_referential_recursive_tree is not None:
            for slot_type, slot_tree in self._slot_tree_lookup.items():
                if slot_type is not enclosing_template_cls:
                    merge_into_slot_tree(
                        enclosing_template_cls,
                        slot_type,
                        slot_tree,
                        self_referential_recursive_tree)

        # Right, so the thing is, this might have introduced a bunch of new
        # recursive reference loops, and those might be buried within several
        # layers of pending trees and stuff. This isn't suuuuper performance
        # critical, so by far the easiest thing to do is just rebuild the
        # tree in its entirety.
        # Note: I suppose in theory we could maybe just do the same as we're
        # doing for the normal slot trees, and just constantly re-merge any
        # self_referential_recursive_tree?
        self._dynamic_class_slot_tree = gather_dynamic_class_slots(
            enclosing_template_cls,
            set(self.dynamic_class_slot_names),
            self._slot_tree_lookup)

    def stringify_all(self) -> str:
        """This is a debug method that creates a prettified string
        version of the entire slot tree lookup (pending and concrete).
        """
        to_join = []
        to_join.append('Resolved (concrete) slots:')
        for template_class, root_node in self._slot_tree_lookup.items():
            to_join.append(f'  {template_class}')
            to_join.append(root_node.stringify(depth=1))

        to_join.append('Pending (forward reference) slots:')
        for ref_lookup_key, container in self._pending_ref_lookup.items():
            to_join.append(f'  {ref_lookup_key}')
            to_join.append(container.pending_root_node.stringify(depth=1))

        return '\n'.join(to_join)



from __future__ import annotations

import itertools
import logging
from collections.abc import Callable
from collections.abc import Collection
from collections.abc import Hashable
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Mapping
from collections.abc import Sequence
from dataclasses import KW_ONLY
from dataclasses import dataclass
from dataclasses import field
from functools import partial
from functools import singledispatch
from typing import NamedTuple
from typing import cast
from typing import overload

from templatey._bootstrapping import EMPTY_TEMPLATE_INSTANCE
from templatey._bootstrapping import EMPTY_TEMPLATE_XABLE
from templatey._provenance import Provenance
from templatey._provenance import ProvenanceNode
from templatey._signature import TemplateSignature
from templatey._slot_tree import SlotTreeNode
from templatey._slot_tree import extract_dynamic_class_slot_types
from templatey._types import InterfaceAnnotationFlavor
from templatey._types import TemplateClass
from templatey._types import TemplateIntersectable
from templatey._types import TemplateParamsInstance
from templatey._types import is_template_instance
from templatey.exceptions import MismatchedTemplateSignature
from templatey.exceptions import TemplateFunctionFailure
from templatey.parser import InterpolatedContent
from templatey.parser import InterpolatedFunctionCall
from templatey.parser import InterpolatedSlot
from templatey.parser import InterpolatedVariable
from templatey.parser import InterpolationConfig
from templatey.parser import ParsedTemplateResource
from templatey.parser import TemplateInstanceContentRef
from templatey.parser import TemplateInstanceDataRef
from templatey.parser import TemplateInstanceVariableRef
from templatey.templates import ComplexContent
from templatey.templates import InjectedValue
from templatey.templates import TemplateConfig

logger = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class FuncExecutionRequest:
    name: str
    args: Iterable[object]
    kwargs: Mapping[str, object]
    result_key: _PrecallCacheKey
    provenance: Provenance


@dataclass(frozen=True, slots=True)
class FuncExecutionResult:
    # Note: must match signature from TemplateFunction!
    name: str
    retval: Sequence[str | TemplateParamsInstance | InjectedValue] | None
    exc: Exception | None

    def filter_injectables(self) -> Iterable[TemplateParamsInstance]:
        if self.retval is not None:
            for item in self.retval:
                if is_template_instance(item):
                    # This doesn't fully work; because of the missing
                    # intersection type, there's nothing linking the xable that
                    # this checks for with the params instance that the type
                    # expects us to yield back
                    yield item  # type: ignore


@dataclass(slots=True)
class RenderEnvRequest:
    to_load: Collection[type[TemplateParamsInstance]]
    to_execute: Collection[FuncExecutionRequest]
    error_collector: list[Exception]

    # These store results; we're adding them inplace instead of needing to
    # merge them later on
    results_loaded: dict[type[TemplateParamsInstance], ParsedTemplateResource]
    results_executed: dict[_PrecallCacheKey, FuncExecutionResult]


# Yes, this is a larger function than it should be.
# But function calls in python are slow, and we actually kinda care about
# performance here.
# TODO: look into solutions that would allow for inlining functions, so you
# could carve this up into separate functions. Can this be done without import
# hooks?
def render_driver(  # noqa: C901, PLR0912, PLR0915
        template_instance: TemplateParamsInstance,
        output: list[str],
        error_collector: list[Exception]
        ) -> Iterable[RenderEnvRequest]:
    """This is a shared method for driving rendering, used by both async
    and sync renderers. It mutates the output list inplace, and yields
    back batched requests for the render environment.
    """
    context = _RenderContext(
        template_preload={},
        function_precall={},
        error_collector=error_collector)
    yield from context.prepopulate(template_instance)
    template_xable = cast(TemplateIntersectable, template_instance)
    root_template_preload = context.template_preload[type(template_instance)]
    render_stack: list[_RenderStackFrame] = [
        _RenderStackFrame(
            parts=root_template_preload.parts,
            part_count=root_template_preload.part_count,
            config=template_xable._templatey_config,
            signature=template_xable._templatey_signature,
            provenance=Provenance((
                ProvenanceNode(
                    encloser_slot_key='',
                    encloser_slot_index=-1,
                    instance_id=id(template_instance),
                    instance=template_instance),)),
            instance=template_instance,
            prerenderers=template_xable._templatey_prerenderers)]

    while render_stack:
        # Quick note on the following code: you might think these isinstance
        # calls are slow, and that you could speed things up by, say, using
        # some kind of sentinel value in a slot. This is incorrect! (At least
        # with 3.13). Isinstance is actually the fastest (and clearest) way
        # to do it.
        try:
            render_frame = render_stack[-1]
            if render_frame.exhausted:
                render_stack.pop()
                continue

            next_part = render_frame.parts[render_frame.part_index]

            # Strings are hardest to deal with because they're containers, so
            # just get that out of the way first
            if isinstance(next_part, str):
                render_frame.part_index += 1
                output.append(next_part)

            elif isinstance(next_part, InterpolatedVariable):
                render_frame.part_index += 1
                unescaped_vars = _ParamLookup(
                    provenance=render_frame.provenance,
                    template_preload=context.template_preload,
                    param_flavor=InterfaceAnnotationFlavor.VARIABLE,
                    error_collector=error_collector,
                    placeholder_on_error='')

                raw_val = unescaped_vars[next_part.name]
                prerenderer = getattr(
                    render_frame.prerenderers, next_part.name, None)
                if prerenderer is None:
                    prerender = raw_val
                else:
                    prerender = cast(str | None, prerenderer(raw_val))

                if prerender is not None:
                    unescaped_val = _apply_format(
                        prerender,
                        next_part.config)
                    escaped_val = render_frame.config.variable_escaper(
                        unescaped_val)
                    # Note that variable interpolations don't support affixes!
                    output.append(escaped_val)

            elif isinstance(next_part, InterpolatedContent):
                render_frame.part_index += 1
                unverified_content = _ParamLookup(
                    provenance=render_frame.provenance,
                    template_preload=context.template_preload,
                    param_flavor=InterfaceAnnotationFlavor.CONTENT,
                    error_collector=error_collector,
                    placeholder_on_error='')
                val_from_params = unverified_content[next_part.name]

                if isinstance(val_from_params, ComplexContent):
                    output.extend(_render_complex_content(
                        val_from_params,
                        _ParamLookup(
                            provenance=render_frame.provenance,
                            template_preload=context.template_preload,
                            param_flavor=InterfaceAnnotationFlavor.VARIABLE,
                            error_collector=error_collector,
                            placeholder_on_error=''),
                        render_frame.config,
                        next_part.config,
                        render_frame.prerenderers,
                        error_collector))

                else:
                    prerenderer = getattr(
                        render_frame.prerenderers, next_part.name, None)
                    if prerenderer is None:
                        prerender = val_from_params
                    else:
                        prerender = cast(
                            str | None, prerenderer(val_from_params))

                    # As usual, values of None are omitted
                    if prerender is not None:
                        formatted_val = _apply_format(
                            prerender,
                            next_part.config)
                        render_frame.config.content_verifier(formatted_val)
                        output.extend(
                            next_part.config.apply_affix(formatted_val))

            # Slots get a little bit more complicated, but the general idea is
            # to append a stack frame for each depth level. We maintain the
            # state of which instance we're on within the stack frame.
            elif isinstance(next_part, InterpolatedSlot):
                slot_instance_index = render_frame.slot_instance_index
                if slot_instance_index == 0:
                    slot_instances = getattr(
                        render_frame.instance, next_part.name)
                    slot_instance_count = len(slot_instances)

                    if slot_instance_count > 0:
                        render_frame.slot_instance_count = slot_instance_count
                        render_frame.slot_instances = slot_instances
                    # Note that this is also important to skip prefix/suffix.
                    else:
                        render_frame.part_index += 1
                        continue

                else:
                    # Note that we skip this entirely if the slot instance
                    # count is zero, so by definition, if we hit this branch,
                    # we need a suffix.
                    output.extend(next_part.config.apply_suffix_iter())
                    slot_instance_count = render_frame.slot_instance_count
                    # We've exhausted the instances; reset the state for the
                    # next slot.
                    if render_frame.slot_instance_index >= slot_instance_count:
                        render_frame.slot_instance_index = 0
                        render_frame.slot_instance_count = 0
                        # Note: we're deliberately skipping the slot instances,
                        # because it'll just get overwritten the next time
                        # around, so this is faster (though it doesn't free
                        # memory as quickly)
                        render_frame.part_index += 1
                        continue

                    slot_instances = render_frame.slot_instances

                # Remember: we skip this entirely if the slot instance count
                # is zero.
                output.extend(next_part.config.apply_prefix_iter())
                slot_instance = slot_instances[slot_instance_index]
                # Note that this needs to support both union slot
                # types, and (eventually) dynamic slot types, hence
                # doing this on every iteration instead of
                # precalculating it for the whole interpolated slot
                slot_instance_preload = context.template_preload[
                    type(slot_instance)]
                slot_instance_xable = cast(
                    TemplateIntersectable, slot_instance)
                render_stack.append(
                    _RenderStackFrame(
                        instance=slot_instance,
                        parts=slot_instance_preload.parts,
                        part_count=slot_instance_preload.part_count,
                        config=slot_instance_xable._templatey_config,
                        signature=slot_instance_xable._templatey_signature,
                        provenance=render_frame.provenance.with_appended(
                            ProvenanceNode(
                                encloser_slot_key=next_part.name,
                                encloser_slot_index=slot_instance_index,
                                instance_id=id(slot_instance),
                                instance=slot_instance)),
                        prerenderers=
                            slot_instance_xable._templatey_prerenderers))

                render_frame.slot_instance_index += 1

            elif isinstance(next_part, InterpolatedFunctionCall):
                render_frame.part_index += 1
                execution_result = context.function_precall[
                    _get_precall_cache_key(render_frame.provenance, next_part)]
                nested_render_node = _build_render_frame_for_func_result(
                    render_frame.instance,
                    render_frame.provenance,
                    next_part,
                    execution_result,
                    render_frame.config,
                    error_collector)
                if nested_render_node is not None:
                    render_stack.append(nested_render_node)

            # Similar to slots, but different enough that it warrants a
            # separate approach. Trust me, I tried to unify them, and 1. I
            # never got it fully working, 2. it was a pretty big hack (a
            # virtual slot mechanism with some __getattr__ shenanigans), and
            # 3. it created way more problems than it was worth.
            # More info in the docstring for _InjectedInstanceContainer.
            elif isinstance(next_part, _InjectedInstanceContainer):
                render_frame.part_index += 1
                injected_instance = next_part.instance
                # Note that this needs to support both union slot
                # types, and (eventually) dynamic slot types, hence
                # doing this on every iteration instead of
                # precalculating it for the whole interpolated slot
                injected_instance_preload = context.template_preload[
                    type(injected_instance)]
                injected_instance_xable = cast(
                    TemplateIntersectable, injected_instance)
                render_stack.append(
                    _RenderStackFrame(
                        instance=injected_instance,
                        parts=injected_instance_preload.parts,
                        part_count=injected_instance_preload.part_count,
                        config=injected_instance_xable._templatey_config,
                        signature=injected_instance_xable._templatey_signature,
                        # Note that the correct ``from_injection`` value is
                        # added when creating the current stack frame.
                        provenance=render_frame.provenance.with_appended(
                            ProvenanceNode(
                                encloser_slot_key='',
                                encloser_slot_index=-1,
                                instance_id=id(injected_instance),
                                instance=injected_instance),),
                        prerenderers=
                            injected_instance_xable._templatey_prerenderers))

            else:
                raise TypeError(
                    'Templatey internal error: unknown template part!',
                    next_part)

        # Note: this could be eg a lookup error because of a missing variable.
        # This isn't redundant with the error collection within prepopulation.
        except Exception as exc:
            error_collector.append(exc)


@dataclass(slots=True)
class _InjectedInstanceContainer:
    """When env functions inject instances of templates, the render
    stack gets into a bit of a weird state. We want to avoid needing to
    extend the stack by one for each and every injected template and/or
    stringified value that the function returned. But we can't modify
    the existing parts of a frame; first, it's a sometimes (always?) a
    tuple, but more importantly, it would screw up our entire indexing
    mechanism.

    In theory we could add a "pending injections" sub-stack or something
    to the _RenderStackFrame, but then we'd be stuck checking for that
    during every single iteration of the render driver loop.

    So instead, we do this: wrap the instance into a container, and
    then (as the LAST, and therefore SLOWEST, check in the render driver
    loop) we can check for the container, and have special logic for
    handling them there.
    """
    instance: TemplateParamsInstance


@dataclass(slots=True)
class _RenderStackFrame:
    instance: TemplateParamsInstance
    config: TemplateConfig
    signature: TemplateSignature
    provenance: Provenance
    prerenderers: NamedTuple
    parts: Sequence[
        str
        | InterpolatedSlot
        | InterpolatedContent
        | InterpolatedVariable
        | InterpolatedFunctionCall
        | _InjectedInstanceContainer]

    _: KW_ONLY
    part_count: int
    part_index: int = field(default=0, init=False)
    slot_instance_count: int = field(default=0, init=False)
    slot_instance_index: int = field(default=0, init=False)
    slot_instances: Sequence[TemplateParamsInstance] = field(init=False)

    @property
    def exhausted(self) -> bool:
        # Note that the slot index exhaustion is handled separately, by
        # controlling exactly when the part index is incremented
        return self.part_index >= self.part_count


@dataclass(slots=True)
class _PrepopulationBatch:
    injection_provenance: Provenance | None
    # Note that this might be an injected template!
    local_root_instance: TemplateParamsInstance
    function_backlog: list[FuncExecutionRequest]
    template_backlog: set[TemplateClass]

    def extract_injections(
            self,
            function_precall:
                dict[_PrecallCacheKey, FuncExecutionResult]
            ) -> list[_PrepopulationBatch]:
        result = []
        for exe_req in self.function_backlog:
            result_cache_key = exe_req.result_key
            function_result = function_precall[result_cache_key]

            for injected_template in function_result.filter_injectables():
                injected_xable = cast(
                    TemplateIntersectable, injected_template)

                result.append(_PrepopulationBatch(
                    injection_provenance=exe_req.provenance,
                    local_root_instance=injected_template,
                    function_backlog=[],
                    template_backlog=set(
                        injected_xable._templatey_signature
                        .included_template_classes)))

        return result

    def extract_invocations(
            self,
            template_preload:
                dict[TemplateClass, ParsedTemplateResource],
            error_collector: list[Exception],
            ) -> _PrepopulationBatch | None:
        function_backlog = []
        injection_provenance = self.injection_provenance
        root_instance = self.local_root_instance
        root_xable = cast(TemplateIntersectable, root_instance)
        slot_tree_lookup = root_xable._templatey_signature._slot_tree_lookup

        for template_cls in self.template_backlog:
            abstract_calls = (
                template_preload[template_cls].function_calls)

            if abstract_calls:
                # Remember: for the root template, it might or might not
                # exist in the slot tree, depending on whether or not it has
                # recursion loops. That being said, we want to be
                # defensive, and not assume that BECAUSE it's missing, it's
                # automatically the root instance.
                if template_cls is type(root_instance):
                    slot_tree_root = slot_tree_lookup.get(
                        template_cls,
                        SlotTreeNode(is_terminus=True))
                else:
                    slot_tree_root = slot_tree_lookup[template_cls]

                # Oof. The combinatorics here are brutal.
                for provenance, abstract_call in itertools.product(
                    Provenance.from_slot_tree(
                        root_instance,
                        slot_tree_root,
                        from_injection=injection_provenance),
                    itertools.chain.from_iterable(abstract_calls.values())
                ):
                    unescaped_vars = _ParamLookup(
                        provenance=provenance,
                        template_preload=template_preload,
                        param_flavor=InterfaceAnnotationFlavor.VARIABLE,
                        error_collector=error_collector,
                        placeholder_on_error='')
                    unverified_content = _ParamLookup(
                        provenance=provenance,
                        template_preload=template_preload,
                        param_flavor=InterfaceAnnotationFlavor.CONTENT,
                        error_collector=error_collector,
                        placeholder_on_error='')

                    # Note that the full call signature is **defined**
                    # within the parsed template body, but it may
                    # **reference** vars and/or content within the template
                    # instance.
                    args = _recursively_coerce_func_execution_params(
                        abstract_call.call_args,
                        template_instance=provenance.slotpath[-1].instance,
                        unescaped_vars=unescaped_vars,
                        unverified_content=unverified_content)
                    kwargs = _recursively_coerce_func_execution_params(
                        abstract_call.call_kwargs,
                        template_instance=provenance.slotpath[-1].instance,
                        unescaped_vars=unescaped_vars,
                        unverified_content=unverified_content)

                    if abstract_call.call_args_exp is not None:
                        args = (*args, *cast(
                            Iterable,
                            _recursively_coerce_func_execution_params(
                                abstract_call.call_args_exp,
                                template_instance=
                                    provenance.slotpath[-1].instance,
                                unescaped_vars=unescaped_vars,
                                unverified_content=unverified_content)))

                    if abstract_call.call_kwargs_exp is not None:
                        kwargs.update(cast(
                            Mapping,
                            _recursively_coerce_func_execution_params(
                                abstract_call.call_kwargs_exp,
                                template_instance=
                                    provenance.slotpath[-1].instance,
                                unescaped_vars=unescaped_vars,
                                unverified_content=unverified_content)))

                    result_cache_key = _get_precall_cache_key(
                        provenance, abstract_call)
                    function_backlog.append(
                        FuncExecutionRequest(
                            abstract_call.name,
                            args=args,
                            kwargs=kwargs,
                            result_key=result_cache_key,
                            provenance=provenance))

        if function_backlog:
            return _PrepopulationBatch(
                injection_provenance=injection_provenance,
                local_root_instance=self.local_root_instance,
                function_backlog=function_backlog,
                template_backlog=set())


@dataclass(slots=True)
class _RenderContext:
    template_preload: dict[TemplateClass, ParsedTemplateResource]
    function_precall: dict[_PrecallCacheKey, FuncExecutionResult]
    error_collector: list[Exception]

    def prepopulate(
            self,
            root_template_instance: TemplateParamsInstance
            ) -> Iterable[RenderEnvRequest]:
        """For the passed root template, populates the template_preload
        and function_precall until either all resources have been
        prepared, or it needs help from the render environment.
        """
        root_template_xable = cast(
            TemplateIntersectable, root_template_instance)
        template_preload: \
            dict[TemplateClass, ParsedTemplateResource] = self.template_preload
        function_precall: \
            dict[_PrecallCacheKey, FuncExecutionResult] = self.function_precall

        batches: list[_PrepopulationBatch] = [_PrepopulationBatch(
            injection_provenance=None,
            local_root_instance=root_template_instance,
            function_backlog=[],
            # Note that this always includes the root template class
            template_backlog=set(
                root_template_xable
                ._templatey_signature.included_template_classes))]

        while batches:
            to_load = set()
            to_execute = []
            # Note that it's important that we're modifying the params for
            # the RenderEnvRequest and not the batches, because the batches
            # will determine where we look for additional injections and/or
            # invocations, regardless of whether or not an execution or load
            # was actually required.
            for batch in batches:
                local_root_instance = batch.local_root_instance
                to_load.update(batch.template_backlog)
                to_execute.extend(batch.function_backlog)
                to_load.update(extract_dynamic_class_slot_types(
                    local_root_instance,
                    cast(TemplateIntersectable, local_root_instance)
                        ._templatey_signature._dynamic_class_slot_tree))
            to_load.difference_update(template_preload)

            yield RenderEnvRequest(
                to_load=to_load,
                to_execute=to_execute,
                error_collector=self.error_collector,
                results_loaded=template_preload,
                results_executed=function_precall)

            completed_batches = batches
            batches = []
            # Yes, it's a little awkward to re-iterate over this, but we need
            # to keep track of the injection provenance, and this is the most
            # convenient way to do it.
            # Also note: it's crticial that we're still looking at batches
            # and not to_load/to_execute, since those will skip already-loaded
            # templates, which would result in us skipping needed invocations.
            for completed_batch in completed_batches:
                batches.extend(completed_batch.extract_injections(
                    function_precall))
                next_batch = completed_batch.extract_invocations(
                    template_preload,
                    self.error_collector)
                if next_batch is not None:
                    batches.append(next_batch)


type _PrecallExecutionRequest = tuple[
    Provenance, InterpolatedFunctionCall]
type _PrecallCacheKey = Hashable


def _get_precall_cache_key(
        provenance: Provenance,
        interpolated_call: InterpolatedFunctionCall
        ) -> _PrecallCacheKey:
    """For a particular template instance and interpolated function
    call, creates the hashable cache key to be used for the render
    context.
    """
    # Note that template provenance includes the actual current template
    # instance, so by definition, this encodes exactly which function call
    # is being referenced
    return (provenance, interpolated_call)


def _render_complex_content(
        complex_content: ComplexContent,
        unescaped_vars: _ParamLookup,
        template_config: TemplateConfig,
        interpolation_config: InterpolationConfig,
        prerenderers: NamedTuple,
        error_collector: list[Exception],
        ) -> Iterable[str]:
    try:
        extracted_vars = {
            key: unescaped_vars[key] for key in complex_content.dependencies}
        extracted_prerenderers = {
            key: getattr(prerenderers, key, None)
            for key in complex_content.dependencies}

        for content_segment in complex_content.flatten(
            extracted_vars, interpolation_config, extracted_prerenderers
        ):
            if isinstance(content_segment, InjectedValue):
                raw_val = content_segment.value
                if raw_val is None:
                    continue

                unescaped_val = _apply_format(raw_val, content_segment.config)

                if content_segment.use_variable_escaper:
                    escaped_val = template_config.variable_escaper(
                        unescaped_val)
                else:
                    escaped_val = unescaped_val

                if content_segment.use_content_verifier:
                    template_config.content_verifier(escaped_val)

                yield escaped_val

            # Note: as usual, None values get omitted!
            elif content_segment is not None:
                formatted_val = _apply_format(
                    content_segment, interpolation_config)
                template_config.content_verifier(formatted_val)
                yield formatted_val

    except Exception as exc:
        exc.add_note('Failed to render complex content!')
        error_collector.append(exc)


def _build_render_frame_for_func_result(  # noqa: C901
        enclosing_instance: TemplateParamsInstance,
        enclosing_provenance: Provenance,
        abstract_call: InterpolatedFunctionCall,
        execution_result: FuncExecutionResult,
        template_config: TemplateConfig,
        error_collector: list[Exception]
        ) -> _RenderStackFrame | None:
    """This constructs a _RenderNode for the given execution result and
    returns it (or None, if there was an error).
    """
    injected_templates: list[tuple[int, TemplateParamsInstance]] = []
    resulting_parts: list[str | _InjectedInstanceContainer] = []
    if execution_result.exc is None:
        if execution_result.retval is None:
            raise TypeError(
                'Impossible branch! Malformed func exe result',
                execution_result)

        for index, result_part in enumerate(execution_result.retval):
            if isinstance(result_part, str):
                resulting_parts.append(
                    template_config.variable_escaper(result_part))
            elif isinstance(result_part, InjectedValue):
                resulting_parts.append(
                    _coerce_injected_value(result_part, template_config))
            elif is_template_instance(result_part):
                injected_templates.append((index, result_part))
                # This is just a placeholder; it gets overwritten in
                # _build_render_stack_extension
                resulting_parts.append('')
            else:
                error_collector.append(_capture_traceback(
                    TypeError(
                        'Invalid return from env function!',
                        execution_result, result_part)))

    else:
        if execution_result.retval is not None:
            raise TypeError(
                'Impossible branch! Malformed func exe result',
                execution_result)

        error_collector.append(_capture_traceback(
            TemplateFunctionFailure('Env function raised!'),
            from_exc=execution_result.exc))

    if injected_templates:
        for index, template_instance in injected_templates:
            resulting_parts[index] = _InjectedInstanceContainer(
                template_instance)

        return _RenderStackFrame(
            parts=resulting_parts,
            part_count=len(resulting_parts),
            config=EMPTY_TEMPLATE_XABLE._templatey_config,
            signature=EMPTY_TEMPLATE_XABLE._templatey_signature,
            # Note: keep this empty here, because we need the instance info
            # to match the injected template, and the whole idea here is to
            # avoid a bunch of extraneous stack frames. We'll add in the
            # correct initial node in the render driver code, where we deal
            # with _InjectedInstanceContainers
            provenance=Provenance(from_injection=enclosing_provenance),
            instance=EMPTY_TEMPLATE_INSTANCE,
            prerenderers=EMPTY_TEMPLATE_XABLE._templatey_prerenderers)

    elif resulting_parts:
        return _RenderStackFrame(
            parts=resulting_parts,
            part_count=len(resulting_parts),
            config=EMPTY_TEMPLATE_XABLE._templatey_config,
            signature=EMPTY_TEMPLATE_XABLE._templatey_signature,
            provenance=Provenance(),
            instance=EMPTY_TEMPLATE_INSTANCE,
            prerenderers=EMPTY_TEMPLATE_XABLE._templatey_prerenderers)


# Note: it would be nice if we could get a little more clever with the types
# on this, but having the lookup be passed in as a callable makes it pretty
# awkward
@dataclass(slots=True, init=False)
class _ParamLookup(Mapping[str, object]):
    """This is a highly-performant layer of indirection that avoids most
    dictionary copies, but nonetheless allows us to both have helpful
    error messages, and collect all possible errors into a single
    ExceptionGroup (without short-circuiting on the first error) while
    rendering.
    """
    provenance: Provenance
    error_collector: list[Exception]
    placeholder_on_error: object
    lookup: Callable[[str], object]
    param_flavor: InterfaceAnnotationFlavor

    def __init__(
            self,
            provenance: Provenance,
            template_preload: dict[TemplateClass, ParsedTemplateResource],
            param_flavor: InterfaceAnnotationFlavor,
            error_collector: list[Exception],
            placeholder_on_error: object):
        self.error_collector = error_collector
        self.placeholder_on_error = placeholder_on_error
        self.provenance = provenance
        self.param_flavor = param_flavor

        if param_flavor is InterfaceAnnotationFlavor.CONTENT:
            self.lookup = partial(
                provenance.bind_content,
                template_preload=template_preload)
        elif param_flavor is InterfaceAnnotationFlavor.VARIABLE:
            self.lookup = partial(
                provenance.bind_variable,
                template_preload=template_preload)
        else:
            raise TypeError(
                'Internal templatey error: _ParamLookup not supported with '
                + 'that flavor', param_flavor)

    def __getitem__(self, name: str) -> object:
        try:
            return self.lookup(name)

        except KeyError as exc:
            self.error_collector.append(_capture_traceback(
                MismatchedTemplateSignature(
                    'Template referenced invalid param in a way that was not '
                    + 'caught during template loading. This could indicate '
                    + 'referencing eg a slot as content, content as var, etc. '
                    + 'Or it could indicate an ellipsis being passed in as '
                    + 'the value for a template parameter. Or it could be a '
                    + 'bug in templatey.',
                    self.provenance.slotpath[-1].instance,
                    name),
                from_exc=exc))
            return self.placeholder_on_error

    def __len__(self) -> int:
        # Note: this is going to be less commonly used (presumably) than just
        # getitem (the only external access to this is through complex content
        # flattening), so don't precalculate this during __init__
        template_instance = self.provenance.slotpath[-1].instance
        template_xable = cast(TemplateIntersectable, template_instance)
        if self.param_flavor is InterfaceAnnotationFlavor.CONTENT:
            return len(template_xable._templatey_signature.content_names)
        elif self.param_flavor is InterfaceAnnotationFlavor.VARIABLE:
            return len(template_xable._templatey_signature.var_names)
        else:
            raise TypeError(
                'Internal templatey error: _ParamLookup not supported with '
                + 'that flavor', self.param_flavor)

    def __iter__(self) -> Iterator[str]:
        # Note: this is going to be less commonly used (presumably) than just
        # getitem (the only external access to this is through complex content
        # flattening), so don't precalculate this during __init__
        template_instance = self.provenance.slotpath[-1].instance
        template_xable = cast(TemplateIntersectable, template_instance)
        if self.param_flavor is InterfaceAnnotationFlavor.CONTENT:
            return (
                getattr(template_instance, attr_name)
                for attr_name
                in template_xable._templatey_signature.content_names)
        elif self.param_flavor is InterfaceAnnotationFlavor.VARIABLE:
            return (
                getattr(template_instance, attr_name)
                for attr_name
                in template_xable._templatey_signature.var_names)
        else:
            raise TypeError(
                'Internal templatey error: _ParamLookup not supported with '
                + 'that flavor', self.param_flavor)


@overload
def _recursively_coerce_func_execution_params(
        param_value: str,
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> str: ...
@overload
def _recursively_coerce_func_execution_params[K: object, V: object](
        param_value: Mapping[K, V],
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> dict[K, V]: ...
@overload
def _recursively_coerce_func_execution_params[T: object](
        param_value: list[T] | tuple[T],
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> tuple[T]: ...
@overload
def _recursively_coerce_func_execution_params(
        param_value: TemplateInstanceContentRef | TemplateInstanceVariableRef,
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> object: ...
@overload
def _recursively_coerce_func_execution_params[T: object](
        param_value: T,
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> T: ...
@singledispatch
def _recursively_coerce_func_execution_params(
        # Note: singledispatch doesn't support type vars
        param_value: object,
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> object:
    """Templatey templates support references to both content and
    variables as call args/kwargs for environment functions. They also
    support both iterables (lists) and mappings (dicts) as literals
    within the template, each of which can also reference content and
    variables, and might themselves contain iterables or mappings.

    This recursively walks the passed execution params, converting all
    of the content or variable references to their values. If the passed
    value was a container, it creates a new copy of the container with
    the references replaced. Otherwise, it simple returns the passed
    value.

    This, the trivial case, handles any situation where the passed
    param value was a plain object.
    """
    return param_value


# Note: I think there might be a bug in pyright re: singledispatch vs overloads
@_recursively_coerce_func_execution_params.register  # type: ignore
def _(
        # Note: singledispatch doesn't support type vars
        param_value: list | tuple | dict,
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> tuple | dict:
    """Again, in the container case, we want to create a new copy of
    the container, replacing its values with the recursive call.
    Note that the keys in nested dictionaries cannot be references,
    only the values.
    """
    if isinstance(param_value, dict):
        return {
            contained_key: _recursively_coerce_func_execution_params(
                contained_value,
                template_instance=template_instance,
                unescaped_vars=unescaped_vars,
                unverified_content=unverified_content)
            for contained_key, contained_value in param_value.items()}

    else:
        return tuple(
            _recursively_coerce_func_execution_params(
                contained_value,
                template_instance=template_instance,
                unescaped_vars=unescaped_vars,
                unverified_content=unverified_content)
            for contained_value in param_value)


# Note: I think there might be a bug in pyright re: singledispatch vs overloads
@_recursively_coerce_func_execution_params.register  # type: ignore
def _(
        # Note: singledispatch doesn't support type vars
        param_value: str,
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> str:
    """We need to be careful here to supply a MORE SPECIFIC dispatch
    type than container for strings, since they are technically also
    containers. Bleh.
    """
    return param_value


# Note: I think there might be a bug in pyright re: singledispatch vs overloads
@_recursively_coerce_func_execution_params.register  # type: ignore
def _(
        param_value: TemplateInstanceContentRef,
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> object:
    """Nested content references need to be retrieved from the
    unverified content. Note that this (along with the nested variable
    references) are the whole reason we're doing this execution params
    coercion in the first place.
    """
    return unverified_content[param_value.name]


# Note: I think there might be a bug in pyright re: singledispatch vs overloads
@_recursively_coerce_func_execution_params.register  # type: ignore
def _(
        param_value: TemplateInstanceVariableRef,
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> object:
    """Nested variable references need to be retrieved from the
    unescaped vars. Note that this (along with the nested content
    references) are the whole reason we're doing this execution params
    coercion in the first place.
    """
    return unescaped_vars[param_value.name]


# Note: I think there might be a bug in pyright re: singledispatch vs overloads
@_recursively_coerce_func_execution_params.register  # type: ignore
def _(
        param_value: TemplateInstanceDataRef,
        *,
        template_instance: TemplateParamsInstance,
        unescaped_vars: _ParamLookup,
        unverified_content: _ParamLookup
        ) -> object:
    """Nested variable references need to be retrieved from the
    unescaped vars. Note that this (along with the nested content
    references) are the whole reason we're doing this execution params
    coercion in the first place.
    """
    return getattr(template_instance, param_value.name)


def _apply_format(raw_value, config: InterpolationConfig) -> str:
    """For both interpolated variables and injected values, we allow
    format specs and conversions to be supplied. We need to actually
    apply these, but the stdlib doesn't really give us a good way of
    doing that. So this is how we do that instead.
    """
    # hot path go fast
    if config is None or config.fmt is None:
        # Note: yes, strings can be formatted with eg padding, but we literally
        # just checked to make sure that there was no format spec, so format
        # would have nothing to do here!
        if isinstance(raw_value, str):
            formatted_value = raw_value
        else:
            formatted_value = format(raw_value)

    else:
        formatted_value = format(raw_value, config.fmt)

    return formatted_value


def _capture_traceback[E: Exception](
        exc: E,
        from_exc: Exception | None = None) -> E:
    """This is a little bit hacky, but it allows us to capture the
    traceback of the exception we want to "raise" but then collect into
    an ExceptionGroup at the end of the rendering cycle. It does pollute
    the traceback with one extra stack level, but the important thing
    is to capture the upstream context for the error, and that it will
    do just fine.

    There's almost certainly a better way of doing this, probably using
    traceback from the stdlib. But this is quicker to code, and that's
    my current priority. Gracefulness can come later!
    """
    try:
        if from_exc is None:
            raise exc
        else:
            raise exc from from_exc

    except type(exc) as exc_with_traceback:
        return exc_with_traceback


def _coerce_injected_value(
        injected_value: InjectedValue,
        template_config: TemplateConfig
        ) -> str:
    """InjectedValue instances are used within the return value of
    environment functions and complex content to indicate that the
    result should be sourced from the variables and/or the content of
    the current render call. This function is responsible for converting
    the ``InjectedValue`` instance into the final resulting string to
    render.
    """
    unescaped_value = _apply_format(
        injected_value.value,
        injected_value.config)

    if injected_value.use_variable_escaper:
        escapish_value = template_config.variable_escaper(unescaped_value)
    else:
        escapish_value = unescaped_value

    if injected_value.use_content_verifier:
        template_config.content_verifier(escapish_value)

    return escapish_value



from __future__ import annotations

import functools
import inspect
import logging
import re
import sys
import typing
from collections import ChainMap
from collections import namedtuple
from collections.abc import Callable
from collections.abc import Collection
from collections.abc import Iterable
from collections.abc import Mapping
from collections.abc import Sequence
from dataclasses import _MISSING_TYPE
from dataclasses import Field
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from textwrap import dedent
from types import FrameType
from typing import Annotated
from typing import Any
from typing import Literal
from typing import Protocol
from typing import dataclass_transform
from typing import overload

from docnote import ClcNote

from templatey._forwardrefs import ForwardRefGeneratingNamespaceLookup
from templatey._forwardrefs import ForwardRefLookupKey
from templatey._forwardrefs import extract_frame_scope_id
from templatey._forwardrefs import resolve_forward_references
from templatey._signature import TemplateSignature
from templatey._types import Content
from templatey._types import DynamicClassSlot
from templatey._types import InterfaceAnnotationFlavor
from templatey._types import Slot
from templatey._types import Var
from templatey.interpolators import NamedInterpolator
from templatey.parser import InterpolationConfig
from templatey.parser import LiteralTemplateString
from templatey.parser import TemplateInstanceContentRef
from templatey.parser import TemplateInstanceDataRef
from templatey.parser import TemplateInstanceVariableRef

logger = logging.getLogger(__name__)


class VariableEscaper(Protocol):

    def __call__(self, value: str) -> str:
        """Variable escaper functions accept a single positional
        argument: the value of the variable to escape. It then does any
        required escaping and returns the final string.
        """
        ...


class ContentVerifier(Protocol):

    def __call__(self, value: str) -> Literal[True]:
        """Content verifier functions accept a single positional
        argument: the value of the content to verify. It does any
        verification, and then returns True if the content was okay,
        or raises BlockedContentValue if the content was not acceptable.

        Note that we raise instead of trying to escape for two reasons:
        1.. We don't really know what to replace it with. This is also
            true with variables, but:
        2.. We expect that content is coming from -- if not trusted,
            then at least authoritative -- sources, and therefore, we
            should fail loudly, because it gives the author a chance to
            correct the problem before it becomes user-facing.
        """
        ...


class InterpolationPrerenderer(Protocol):

    def __call__(
            self,
            value: Annotated[
                object | None,
                ClcNote(
                    '''The value of the variable or content. A value of
                    ``None`` indicates that the value is intended to be
                    omitted, but you may still provide a fallback
                    instead.
                    ''')]
            ) -> str | None:
        """Interpolation prerenderers give you a chance to modify the
        rendered result of a particular content or variable value, omit
        it entirely, or provide a fallback for missing values.

        Prerenderers are applied before formatting, escaping, and
        verification, and the result of the prerenderer is used to
        determine whether or not the value should be included in the
        result. If your prerenderer returns ``None``, the parameter will
        be completely omitted (including any prefix or suffix).
        """
        ...


# The following is adapted directly from typeshed. We did some formatting
# updates, and inserted our prerenderer param.
if sys.version_info >= (3, 14):
    @overload
    def param[_T](
            *,
            default: _T,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            init: bool = True,
            repr: bool = True,
            hash: bool | None = None,
            compare: bool = True,
            metadata: Mapping[Any, Any] | None = None,
            kw_only: bool | Literal[_MISSING_TYPE.MISSING] = ...,
            doc: str | None = None,
            prerenderer: InterpolationPrerenderer | None = None,
            ) -> _T: ...
    @overload
    def param[_T](
            *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Callable[[], _T],
            init: bool = True,
            repr: bool = True,
            hash: bool | None = None,
            compare: bool = True,
            metadata: Mapping[Any, Any] | None = None,
            kw_only: bool | Literal[_MISSING_TYPE.MISSING] = ...,
            doc: str | None = None,
            prerenderer: InterpolationPrerenderer | None = None,
            ) -> _T: ...
    @overload
    def param[_T](
            *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            init: bool = True,
            repr: bool = True,
            hash: bool | None = None,
            compare: bool = True,
            metadata: Mapping[Any, Any] | None = None,
            kw_only: bool | Literal[_MISSING_TYPE.MISSING] = ...,
            doc: str | None = None,
            prerenderer: InterpolationPrerenderer | None = None,
            ) -> Any: ...

# This is technically only valid for >=3.10, but we require that anyways
else:
    @overload
    def param[_T](
            *,
            default: _T,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            init: bool = True,
            repr: bool = True,
            hash: bool | None = None,
            compare: bool = True,
            metadata: Mapping[Any, Any] | None = None,
            kw_only: bool | Literal[_MISSING_TYPE.MISSING] = ...,
            prerenderer: InterpolationPrerenderer | None = None,
            ) -> _T: ...
    @overload
    def param[_T](
            *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Callable[[], _T],
            init: bool = True,
            repr: bool = True,
            hash: bool | None = None,
            compare: bool = True,
            metadata: Mapping[Any, Any] | None = None,
            kw_only: bool | Literal[_MISSING_TYPE.MISSING] = ...,
            prerenderer: InterpolationPrerenderer | None = None,
            ) -> _T: ...
    @overload
    def param[_T](
            *,
            default: Literal[_MISSING_TYPE.MISSING] = ...,
            default_factory: Literal[_MISSING_TYPE.MISSING] = ...,
            init: bool = True,
            repr: bool = True,
            hash: bool | None = None,
            compare: bool = True,
            metadata: Mapping[Any, Any] | None = None,
            kw_only: bool | Literal[_MISSING_TYPE.MISSING] = ...,
            prerenderer: InterpolationPrerenderer | None = None,
            ) -> Any: ...


def param(
        *,
        prerenderer: InterpolationPrerenderer | None = None,
        metadata: Mapping[Any, Any] | None = None,
        **field_kwargs):
    if metadata is None:
        metadata = {'templatey.prerenderer': prerenderer}

    else:
        metadata = {
            **metadata,
            'templatey.prerenderer': prerenderer}

    return field(**field_kwargs, metadata=metadata)


@dataclass_transform(field_specifiers=(param, field, Field))
def template[T: type](  # noqa: PLR0913
        config: TemplateConfig,
        template_resource_locator: object,
        /, *,
        init: bool = True,
        repr: bool = True,  # noqa: A002
        eq: bool = True,
        order: bool = False,
        unsafe_hash: bool = False,
        frozen: bool = False,
        match_args: bool = True,
        kw_only: bool = False,
        slots: bool = True,
        weakref_slot: bool = False,
        segment_modifiers: Annotated[
                Sequence[SegmentModifier] | None,
                ClcNote('''An ordered sequence of ``SegmentModifier`` instances
                    to apply to every literal string segment of the loaded
                    template text.

                    The modifiers will be applied in the order that they are
                    declared. The first modifier to match the segment will
                    short-circuit the remaining modifiers, regardless of
                    whether or not it applies any changes.
                    ''')
            ] = None
        ) -> Callable[[T], T]:
    """This both transforms the decorated class into a stdlib dataclass
    and declares it as a templatey template.

    **Note that unlike the stdlib dataclass decorator, this defaults to
    ``slots=True``.** If you find yourself having problems with
    metaclasses and/or subclassing, you can disable this by passing
    ``slots=False``. Generally speaking, though, this provides a
    free performance benefit. **If weakref support is required, be sure
    to pass ``weakref_slot=True``.
    """
    if segment_modifiers is None:
        segment_modifiers = []

    return functools.partial(
        make_template_definition,
        dataclass_kwargs={
            'init': init,
            'repr': repr,
            'eq': eq,
            'order': order,
            'unsafe_hash': unsafe_hash,
            'frozen': frozen,
            'match_args': match_args,
            'kw_only': kw_only,
            'slots': slots,
            'weakref_slot': weakref_slot
        },
        template_resource_locator=template_resource_locator,
        template_config=config,
        segment_modifiers=segment_modifiers)


@dataclass(frozen=True)
class TemplateConfig[T: type, L: object]:
    interpolator: Annotated[
        NamedInterpolator,
        ClcNote(
            '''The interpolator determines what characters are used for
            performing interpolations within the template. They can be
            escaped by repeating them, for example ``{{}}`` would be
            a literal ``{}`` with a curly braces interpolator.
            ''')]
    variable_escaper: Annotated[
        VariableEscaper,
        ClcNote(
            '''Variables are always escaped. The variable escaper is
            the callable responsible for performing that escaping. If you
            don't need escaping, there are noop escapers within the prebaked
            template configs that you can use for convenience.
            ''')]
    content_verifier: Annotated[
        ContentVerifier,
        ClcNote(
            '''Content isn't escaped, but it ^^is^^ verified. Content
            verification is a simple process that either succeeds or fails;
            it allows, for example, to allowlist certain HTML tags.
            ''')]


def _extract_template_class_locals() -> dict[str, Any] | None:
    """When templates are created from inside a closure (ex, during
    testing, where this is extremely common), we need access to the
    locals from the closure to resolve type hints. This method relies
    upon ``inspect`` to extract them.

    Note that this can be very sensitive to where, exactly, you put it
    within the templatey code. Always put it as close as possible to
    the public API method, so that the first frame from another module
    coincides with the call to decorate a template class.
    """
    upmodule_frame = _get_first_frame_from_other_module()
    if upmodule_frame is not None:
        return upmodule_frame.f_locals


def _get_first_frame_from_other_module() -> FrameType | None:
    """Both of our closure workarounds require walking up the stack
    until we reach the first frame coming from ^^outside^^ the ~~house~~
    current module. This performs that lookup.

    **Note that this is pretty fragile.** Or, put a different way: it
    does exactly what the function name suggest it does: it finds the
    FIRST frame from another module. That doesn't mean we won't return
    to this module; it doesn't mean it's from the actual client library,
    etc. It just means it's the first frame that isn't from this
    module.
    """
    upstack_frame = inspect.currentframe()
    if upstack_frame is None:
        return None
    else:
        this_module = upstack_module = inspect.getmodule(
            _extract_template_class_locals)
        while upstack_module is this_module:
            if upstack_frame is None:
                return None

            upstack_frame = upstack_frame.f_back
            upstack_module = inspect.getmodule(upstack_frame)

    return upstack_frame


# Yes, this is awkward with a bazillion return statements, but in this case,
# clarity is better than elegance
def _classify_interface_field_flavor(  # noqa: PLR0911
        parent_class_type_hints: dict[str, Any],
        template_field: Field
        ) -> tuple[set[InterfaceAnnotationFlavor], type | None]:
    """For a dataclass field, determines whether it was declared as a
    var, slot, or content.

    If none of the above, returns None.
    """
    # Note that dataclasses don't include the actual type (just a string)
    # when in __future__ mode, so we need to get them from the parent class
    # by calling get_type_hints() on it
    resolved_field_type = parent_class_type_hints[template_field.name]
    anno_origin = typing.get_origin(resolved_field_type)
    if anno_origin is Var:
        nested_type, = typing.get_args(resolved_field_type)
        return {InterfaceAnnotationFlavor.VARIABLE}, nested_type
    elif anno_origin is Slot:
        nested_type, = typing.get_args(resolved_field_type)
        return {InterfaceAnnotationFlavor.SLOT}, nested_type
    elif anno_origin is Content:
        nested_type, = typing.get_args(resolved_field_type)
        return {InterfaceAnnotationFlavor.CONTENT}, nested_type
    elif anno_origin is DynamicClassSlot:
        return {
            InterfaceAnnotationFlavor.SLOT,
            InterfaceAnnotationFlavor.DYNAMIC
        }, None

    # This is all if there's a generic annotation with no parameter passed,
    # ex ``foo: Var`` or ``bar: Slot`` or (presumably) ``baz: list``
    elif anno_origin is None:
        if resolved_field_type is Var:
            return {InterfaceAnnotationFlavor.VARIABLE}, None

        elif resolved_field_type is Content:
            return {InterfaceAnnotationFlavor.CONTENT}, None

        elif resolved_field_type is DynamicClassSlot:
            return {
                InterfaceAnnotationFlavor.SLOT,
                InterfaceAnnotationFlavor.DYNAMIC
            }, None

        elif resolved_field_type is Slot:
            raise TypeError(
                '``Slot`` annotations require a concrete slot class as a '
                + 'parameter!')

        else:
            return {InterfaceAnnotationFlavor.DATA}, None

    # There was a non-generic parameter passed. Therefore, it can only be
    # one thing: template data.
    else:
        return {InterfaceAnnotationFlavor.DATA}, None


@dataclass_transform(field_specifiers=(param, field, Field))
def make_template_definition[T: type](
        cls: T,
        *,
        dataclass_kwargs: dict[str, bool],
        # Note: needs to be understandable by template loader
        template_resource_locator: object,
        template_config: TemplateConfig,
        segment_modifiers: Sequence[SegmentModifier]
        ) -> T:
    """Programmatically creates a template definition. Converts the
    requested class into a dataclass, passing along ``dataclass_kwargs``
    to the dataclass constructor. Then performs some templatey-specific
    bookkeeping. Returns the resulting dataclass.
    """
    cls = dataclass(**dataclass_kwargs)(cls)
    cls._templatey_config = template_config
    cls._templatey_resource_locator = template_resource_locator

    template_module = cls.__module__
    template_scope_id = extract_frame_scope_id()
    template_forward_ref = ForwardRefLookupKey(
        module=template_module,
        name=cls.__name__,
        scope_id=template_scope_id)

    # We're prioritizing the typical case here, where the templates are defined
    # at the module toplevel, and therefore accessible within the module
    # globals. However, if the template is defined within a closure, we might
    # need to walk up the stack until we find a caller that isn't within this
    # file, and then grab its locals.
    try:
        template_type_hints = typing.get_type_hints(cls)
    except NameError as exc:
        logger.info(dedent('''\
            Failed to resolve template type hints on first pass. This could be
            indicative of a bug, or it might occur in normal situations if:
            ++  you're defining the template within a closure. Here, we'll
                attempt to infer the locals via inspect.currentframe, but not
                all platforms support that, which can lead to failures
            ++  the type hint is a forward reference.

            In both cases, we'll wrap the request into a
            ``ForwardRefLookupKey``, which will then hopefully be
            resolved as soon as the forward reference is declared.
            If it's never resolved, however, we will raise whenever
            ``render`` is called.
            '''),
            exc_info=exc)

        # If you're confused, keep reading the comments below.
        forwardref_lookup = ForwardRefGeneratingNamespaceLookup(
            template_module=template_module,
            template_scope_id=template_scope_id)
        # This is the same as the current implementation of get_type_hints
        # in cpython for classes:
        # https://github.com/python/cpython/blob/0045100ccbc3919e8990fa59bc413fe38d21b075/Lib/typing.py#L2325
        template_globals = getattr(
            sys.modules.get(template_module, None), '__dict__', {})

        # Okay, so... this is non-intuitive, to be sure. And if you fail to
        # do this, you'll have really perfidious bugs, because anything that
        # was a valid reference within the temporary namespace created inside
        # the class creation -- **including the class itself** -- will be
        # converted to a forward reference proxy. This would of course cause
        # problems with inter-class references, but it also means that the
        # class itself gets treated as a forward ref instead a concrete slot,
        # and therefore added into the pending forward refs registration. This
        # then gets immediately fixed, except, because the registration is a
        # set, and the iteration order of sets is non-deterministic, the
        # exact, arbitrary order of pending ref resolution can cause a
        # correctly-resolved forward ref for the current class being processed,
        # to then be overwritten by a pending reference, which is then never
        # resolved. Believe me, it's a mess.
        # So to solve this, we need to do two things. First, add the class'
        # __dict__ into the namespace lookups, because that solves any
        # inter-class references. But second, we need to also add the class
        # itself, under its __name__, because this name, though valid and
        # accessible in the temporary class creation namespace, is nonetheless
        # removed from the __dict__ during final class creation.
        # But, ONE FINAL THING: chainmaps need **mutable** mappings. So we
        # have to create a copy of the cls.__dict__ anyways.
        resurrected_clsdef_namespace = {**cls.__dict__, cls.__name__: cls}

        # Note: the whole point of this is to support forward refs, and there's
        # a method to the madness here. We need a way of getting an actual
        # forward ref object. Prior to 3.14 (via PEP 649), we don't have a way
        # to specify to ``get_type_hints`` that we want to have a forward ref
        # object, and it just raises a nameerror that we can't recover from.
        # Our workaround is to inject an object that will lazily create proxy
        # objects for all forward references, but we have to supply this to
        # get_type_hints somehow, which is non-trivial:
        # globalns needs to be strictly a dict, because it gets delegated into
        # ``eval``, which requires one. Which means we can only use the localns
        # to intercept missing forward references. But that, then, means that
        # we need to recover the existing check for the actual globals, since
        # otherwise **all** global names would be overwritten by the forward
        # reference.
        maybe_locals = _extract_template_class_locals()
        if maybe_locals is None:
            prioritized_lookups = (
                # See note above about this
                resurrected_clsdef_namespace,
                template_globals,
                # Fun fact: these aren't included in the other globals!
                __builtins__,
                forwardref_lookup)

        else:
            prioritized_lookups = (
                # See note above about this
                resurrected_clsdef_namespace,
                maybe_locals,
                template_globals,
                # Fun fact: these aren't included in the other globals!
                __builtins__,
                forwardref_lookup)

        # Because of our forward lookup, this will always succeed
        template_type_hints = typing.get_type_hints(
            cls, localns=ChainMap(*prioritized_lookups))

    slots = {}
    dynamic_slots = {}
    vars_ = {}
    content = {}
    data = {}
    prerenderers = {}
    # Note that this ignores initvars, which is what we want
    for template_field in fields(cls):
        field_classification = _classify_interface_field_flavor(
            template_type_hints, template_field)

        # Note: it's not entirely clear to me that this restriction makes
        # sense; I could potentially see MAYBE there being some kind of
        # environment function that could access other attributes from the
        # dataclass? But also, maybe those should be vars? Again, unclear.
        if field_classification is None:
            raise TypeError(
                'Template parameter definitions may only contain variables, '
                + 'slots, and content!')

        else:
            field_flavors, wrapped_type = field_classification

            # A little awkward to effectively just repeat the comparison we did
            # when classifying, but that makes testing easier and control flow
            # clearer
            if InterfaceAnnotationFlavor.VARIABLE in field_flavors:
                dest_lookup = vars_
            elif InterfaceAnnotationFlavor.SLOT in field_flavors:
                if InterfaceAnnotationFlavor.DYNAMIC in field_flavors:
                    dest_lookup = dynamic_slots
                else:
                    dest_lookup = slots
            elif InterfaceAnnotationFlavor.CONTENT in field_flavors:
                dest_lookup = content
            else:
                dest_lookup = data

            dest_lookup[template_field.name] = wrapped_type
            prerenderers[template_field.name] = template_field.metadata.get(
                'templatey.prerenderer')

    cls._templatey_signature = TemplateSignature.new(
        template_cls=cls,
        data=data,
        slots=slots,
        dynamic_class_slot_names=set(dynamic_slots),
        vars_=vars_,
        content=content,
        forward_ref_lookup_key=template_forward_ref)
    converter_cls = namedtuple('TemplateyConverters', tuple(prerenderers))
    cls._templatey_prerenderers = converter_cls(**prerenderers)
    cls._templatey_segment_modifiers = tuple(segment_modifiers)

    # Note: this needs to be the absolute last thing, because we need to fully
    # satisfy the intersectable interface before we can call it.
    resolve_forward_references(cls)
    return cls


@dataclass(frozen=True, slots=True)
class InjectedValue:
    """This is used by environment functions and complex content to
    indicate that a value is being injected into the template. Use it
    instead of a bare string to preserve an existing interpolation
    config, or to indicate whether verification and/or escaping should
    be applied to the value after conversion to a string.

    Note that, if both are defined, the variable escaper will be called
    first, before the content verifier.
    """
    value: object

    config: InterpolationConfig = field(default_factory=InterpolationConfig)
    use_content_verifier: bool = False
    use_variable_escaper: bool = True

    def __post_init__(self):
        if self.config.prefix is not None or self.config.suffix is not None:
            raise ValueError(
                'Injected values cannot have prefixes nor suffixes. If you '
                + 'need similar behavior, simply add the affix(es) to the '
                + 'iterable returned by the complex content or env function.')


class _ComplexContentBase(Protocol):

    def flatten(
            self,
            dependencies: Annotated[
                Mapping[str, object],
                ClcNote(
                    '''The values of the variables declared as dependencies
                    in the constructor are passed to the call to ``flatten``
                    during rendering.
                    ''')],
            config: Annotated[
                InterpolationConfig,
                ClcNote(
                    '''The interpolation configuration of the content
                    interpolation that the complex content is a member
                    of. Note that neither prefix nor suffix can be passed
                    on to an ``InjectedValue``; they must be manually included
                    in the return value if desired.
                    ''')],
            prerenderers: Annotated[
                Mapping[str, InterpolationPrerenderer | None],
                ClcNote(
                    '''If a prerenderer is defined on a dependency variable,
                    it will be included here; otherwise, the value will be
                    set to ``None``.
                    ''')],
            ) -> Iterable[object | InjectedValue]:
        """Implement this for any instance of complex content.

        First, do whatever content modification you need to, based on
        the dependency variables declared in the constructor. Then,
        if needed, merge in the variable itself using an
        ``InjectedValue``, configuring it as appropriate.

        **Note that the parent interpolation config will be ignored by
        all strings returned by flattening individually.** So if, for
        example, you included a prefix in the content interpolation
        within the template itself, and then passed a ``ComplexContent``
        instance to the template instance, the prefix would be ignored
        completely (unless you do something with it in ``flatten``).

        **Also note that you are responsible for calling the dependency
        variable's ``InterpolationPrerenderer``. directly,** within your
        implementation of ``flatten``. This affords you the option to
        skip it if desired.

        > Example: noun quantity
        __embed__: 'code/python'
            class NaivePluralContent(ComplexContent):

                def flatten(
                        self,
                        dependencies: Mapping[str, object],
                        config: InterpolationConfig,
                        prerenderers:
                            Mapping[str, InterpolationPrerenderer | None],
                        ) -> Iterable[str | InjectedValue]:
                    \"""Pluralizes the name of the provided dependency.
                    For example, ``{'widget': 1}`` will be rendered as
                    "1 widget", but ``{'widget': 2}`` will be rendered as
                    "2 widgets".
                    \"""

                    # Assume only 1 dependency
                    name, value = next(iter(dependencies.items()))

                    if 0 <= value <= 1:
                        return (
                            InjectedValue(
                                value,
                                # This assumes no prefix/suffix
                                config=config,
                                use_content_verifier=False,
                                use_variable_escaper=True),
                            ' ',
                            name)

                    else:
                        return (
                            InjectedValue(
                                value,
                                # This assumes no prefix/suffix
                                config=config,
                                use_content_verifier=False,
                                use_variable_escaper=True),
                            ' ',
                            name,
                            's')
        """
        ...


@dataclass(slots=True, kw_only=True)
class ComplexContent(_ComplexContentBase):
    """Sometimes content isn't as simple as a ``string``. For example,
    content might include variable interpolations. Or you might need
    to modify the content slightly based on the variables -- for
    example, to get subject/verb alignment based on a number, gender
    alignment based on a pronoun, or whatever. ComplexContent gives
    you an escape hatch to do this: simply pass a ComplexContent
    instance as a value instead of a string.
    """

    dependencies: Annotated[
        Collection[str],
        ClcNote(
            '''Complex content dependencies are the **variable** names
            that a piece of complex content depends on. These will be
            passed to the implemented ``flatten`` function during
            rendering.
            ''')]


@dataclass(slots=True, frozen=True)
class SegmentModifier:
    """Segment modifiers are a particularly powerful tool for reducing
    the verbosity of the actual template text. They allow you to modify
    literal string segments that match a particular pattern, replacing
    that pattern with a different segment sequence.

    For example, you could have a modifier that replaces every newline
    in the segment with a call to an ``add_indentation`` environment
    function:

    Note that if you want access to any part of the matched pattern,
    you must include it in a capture group:
    """
    pattern: re.Pattern
    modifier: Callable[
        [SegmentModifierMatch],
        Sequence[
            EnvFuncInvocationRef
            | TemplateInstanceContentRef
            | TemplateInstanceVariableRef
            | str]]

    def apply_and_flatten(
            self,
            literal_template_string: LiteralTemplateString
            ) -> tuple[
                bool,
                list[
                    str
                    | EnvFuncInvocationRef
                    | TemplateInstanceContentRef
                    | TemplateInstanceVariableRef]]:
        """This takes a literal template string segment and applies all
        modifiers. This gets it ready to be converted into the actual
        ParsedTemplateResource parts we need, but doesn't yet perform
        the final conversion.

        Returns a tuple of [had_matches, segments_after_modification]
        """
        after_modification: list[
            str
            | EnvFuncInvocationRef
            | TemplateInstanceContentRef
            | TemplateInstanceVariableRef] = []
        had_matches = False
        splits = self.pattern.split(literal_template_string)
        for split_segment_or_modification in SegmentModifierMatch.from_splits(
            self.pattern,
            splits
        ):
            if isinstance(split_segment_or_modification, str):
                # Here's where we filter out any empty strings. This also helps
                # us re-normalize the output from re.split, so that we recover
                # from the land of "let's use an empty string as a placeholder
                # for anything we did implicitly"
                if split_segment_or_modification:
                    after_modification.append(split_segment_or_modification)
            else:
                had_matches = True
                after_modification.extend(
                    self.modifier(split_segment_or_modification))

        return had_matches, after_modification


@dataclass(slots=True, kw_only=True)
class SegmentModifierMatch:
    captures: Annotated[
        list[str],
        ClcNote(
            '''The captures in a segment modifier match correspond to
            the capture groups in the ``SegmentModifier`` pattern that
            resulted in the match. If the pattern had zero match groups,
            it will be an empty list. If it had one match group, it will
            be a single-item list. Etc.

            The ordering will be the same as the ordering of the groups
            in the original pattern. Named groups are not supported.
            ''')]

    @classmethod
    def from_splits(
            cls,
            src_pattern: re.Pattern,
            splits: list[str]
            ) -> Iterable[SegmentModifierMatch | str]:
        # Note: the +1 is because we need to account for the literal strings
        # in the pattern; otherwise the mod math doesn't work out! This also
        # has the added benefit of meaning we don't need to change behavior
        # between both cases.
        split_type_count = src_pattern.groups + 1

        current_captures = []
        for index, post_split_segment in enumerate(splits):
            # This is always literal text, even if it's an empty string.
            if not index % split_type_count:
                # Note that the zeroth and last index are always literal text,
                # and are never preceeded or followed by a capture. However,
                # we still need to yield the preceeding match when we reach
                # the end.
                if 0 < index:
                    yield cls(captures=current_captures)
                    current_captures = []

                # Note: we're saving the "cull empty strings" bit for later;
                # it makes the logic much cleaner to do one at a time
                yield post_split_segment

            else:
                current_captures.append(post_split_segment)


@dataclass(slots=True, init=False)
class EnvFuncInvocationRef:
    """Used to indicate that a segment modification needs to invoke an
    environment function. Instantiate these like partials:

    > Invocation example
        def my_env_func(foo: str, *, bar: int) -> list[str]:
            ...

        EnvFuncInvocationRef('my_env_func', 'foo', bar=3)
        EnvFuncInvocationRef(
            'my_env_func',
            # TemplateInstanceContentRef, TemplateInstanceDataRef, and
            # TemplateInstanceVariableRef are all supported, including nested
            # within containers
            TemplateInstanceVariableRef('foo'),
            bar=TemplateInstanceDataRef('bar'))

    Note that unfortunately, until/unless higher-kinded type support is
    added to python, these won't be able to type-check correctly.
    """
    name: str
    call_args: tuple[
        object
        | TemplateInstanceContentRef
        | TemplateInstanceDataRef
        | TemplateInstanceVariableRef, ...]
    call_kwargs: dict[
        str,
        object
        | TemplateInstanceContentRef
        | TemplateInstanceDataRef
        | TemplateInstanceVariableRef]

    def __init__(
            self,
            name: str,
            /,
            *call_args:
                object
                | TemplateInstanceContentRef
                | TemplateInstanceDataRef
                | TemplateInstanceVariableRef,
            **call_kwargs:
                object
                | TemplateInstanceContentRef
                | TemplateInstanceDataRef
                | TemplateInstanceVariableRef):
        self.name = name
        self.call_args = call_args
        self.call_kwargs = call_kwargs



from __future__ import annotations

import operator
from collections.abc import Iterable
from collections.abc import Sequence
from copy import copy
from dataclasses import KW_ONLY
from dataclasses import InitVar
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from typing import Self
from typing import cast
from typing import overload

from templatey._forwardrefs import ForwardRefLookupKey
from templatey._types import TemplateClass
from templatey._types import TemplateIntersectable
from templatey._types import TemplateParamsInstance
from templatey._types import create_templatey_id


# Note: the ordering here is to emphasize the fact that the slot
# name is on the ENCLOSING template, but the slot type is from the
# NESTED template
class SlotTreeRoute[T: SlotTreeNode](tuple[str, TemplateClass, T]):
    """An individual route on the slot tree is defined by the attribute
    name for the slot, the slot type, and the subtree from the slot
    class.

    These are optimized for the non-union case. Traversing the slot tree
    with union types will result in a bunch of unnecessary comparisons
    against slot names of different slot types.

    Note that slot tree routes always have a concrete slot name and slot
    type, regardless of whether they're in a pending or concrete tree.
    The reason is simple: in a pending tree, all of the pending classes
    are dead-end nodes, and define their insertion points using just
    the string of the slot name, and nothing else.
    """
    @classmethod
    def new(
            cls,
            slot_name: str,
            slot_type: TemplateClass,
            subtree: T,
            ) -> SlotTreeRoute[T]:
        return cls((slot_name, slot_type, subtree))

    @property
    def subtree(self) -> T:
        """This is slower than directly accessing the tuple values, but
        it makes for clearer code during tree building, where
        performance isn't quite so critical.
        """
        return self[2]

    @property
    def slot_path(self) -> tuple[str, TemplateClass]:
        return self[0:2]


@dataclass(slots=True)
class SlotTreeNode[T: SlotTreeNode](list[SlotTreeRoute[T]]):
    """The purpose of the slot tree is to precalculate what sequences of
    getattr() calls we need to traverse to arrive at every instance of a
    particular slot type for a given template, including all nested
    templates.

    **These are optimized for rendering, not for template declaration.**
    Also note that these are optimized for slots that are not declared
    as type unions; type unions will result in a number of unnecessary
    comparisons against the routes of the other slot types in the union.

    The reason this is useful is for batching during rendering. This is
    important for function calls: it allows us to pre-execute all env
    func calls for a template before we start rendering it. In the
    future, it will also serve the same role for discovering the actual
    template types for dynamic slots, allowing us to load the needed
    template types in advance.

    An individual node on the slot tree is a list of all possible
    attribute names (as ``SlotTreeRoute``s) that a particular search
    pass needs to check for a given instance. Note that **all** of the
    attributes must be searched -- hence using an iteration-optimized
    list instead of a mapping.
    """
    routes: InitVar[Iterable[SlotTreeRoute[T]] | None] = None

    _: KW_ONLY
    # We use this to limit the number of entries we need in the transmogrifier
    # lookup during tree merging/copying
    is_recursive: bool = False
    # We use this to differentiate between dissimilar unions, one which
    # continues on to a subtree, and one which ends here with the target
    # instance
    is_terminus: bool = False

    id_: int = field(default_factory=create_templatey_id)

    # We use this to make the logic cleaner when merging trees, but we want
    # the faster performance of the tuple when actually traversing the tree
    _index_by_slot_path: dict[tuple[str, TemplateClass], int] = field(
            init=False, repr=False)

    def __post_init__(
            self,
            routes: Iterable[SlotTreeRoute[T]] | None):
        # Explicit instead of super because... idunno, we're breaking things
        # somehow
        if routes is None:
            list.__init__(self)
        else:
            list.__init__(self, routes)

        self._index_by_slot_path = {
            route.slot_path: index for index, route in enumerate(self)}

    def empty_clone(
            self,
            *,
            fields_to_skip: set[str] | frozenset[str] = frozenset()
            ) -> Self:
        """This creates a clone of the node without any routes. Useful
        for merging and copying, where you need to do some manual
        transform of the content.

        Note that this is almost the same as dataclasses.replace, with
        the exception that we create shallow copies of attributes
        instead of preserving them.
        """
        kwargs = {}
        for dc_field in fields(self):
            if dc_field.init and dc_field.name not in fields_to_skip:
                # Note: the copy here is important for any mutable values,
                # notably the insertion_slot_names.
                kwargs[dc_field.name] = copy(getattr(self, dc_field.name))

        return type(self)(**kwargs)

    def merge_fields_only(
            self,
            other: SlotTreeNode,
            *,
            fields_to_skip: set[str] | frozenset[str] = frozenset()):
        """Updates the current node, merging in all non-init field
        values from other, using |=. Only merges values that exist on
        the current node, allowing for transformation between pending
        and concrete node types.

        Leaves the ID of the current node unchanged.
        """
        # Always skip ``id_`` -- always preserve the original ID!
        fields_to_skip = fields_to_skip | {'id_'}
        missing = object()

        for dc_field in fields(self):
            if dc_field.init and dc_field.name not in fields_to_skip:
                current_value = getattr(self, dc_field.name)
                other_value = getattr(other, dc_field.name, missing)

                if other_value is not missing:
                    setattr(
                        self,
                        dc_field.name,
                        operator.ior(current_value, other_value))

    @property
    def requires_transmogrification(self) -> bool:
        """This determines whether or not copies of the tree require
        some post-processing to make sure that the tree STRUCTURE is
        the same. It is used by both copying and merging trees to make
        sure that the transmogrification lookup is as sparse as
        possible.

        For the base class, we simply wrap ``is_recursive``, but for
        the pending tree derived class, we also check other stuff --
        hence the wrapping.
        """
        return self.is_recursive

    def append(self, route: SlotTreeRoute[T]):
        """This has slightly different semantics to normal list
        appending if you're trying to append a duplicate slot path:
        ++  if it's identical to the existing path (ie, it targets the
            same node), we do nothing
        ++  if it's not identical, we error
        """
        slot_path = (route[0], route[1])
        slot_index = self._index_by_slot_path.get(slot_path)

        if slot_index is not None:
            if self[slot_index].subtree is route.subtree:
                return
            else:
                raise ValueError(
                    'Templatey internal error: attempt to append duplicate '
                    + 'slot name for same slot type, but a different subtree! '
                    + 'Please search for / report issue to github along with '
                    + 'a traceback.')

        list.append(self, route)
        self._index_by_slot_path[slot_path] = len(self) - 1

    def has_route_for(
            self,
            slot_name: str,
            slot_type: TemplateClass
            ) -> bool:
        return (slot_name, slot_type) in self._index_by_slot_path

    def get_route_for(
            self,
            slot_name: str,
            slot_type: TemplateClass
            ) -> SlotTreeRoute[T]:
        return self[self._index_by_slot_path[(slot_name, slot_type)]]

    def rewrite_route_for(
            self,
            slot_name: str,
            slot_type: TemplateClass,
            new_route: SlotTreeRoute[T]
            ) -> None:
        dest_index = self._index_by_slot_path[(slot_name, slot_type)]
        list.__setitem__(self, dest_index, new_route)

    def stringify(
            self,
            *,
            depth=0,
            _encountered_ids: frozenset[int] = frozenset()
            ) -> str:
        """Creates a pretty-print-style string representation of the
        node and all its nested nodes, recursively.
        """
        indentation = '    ' * depth

        to_join = []
        for dc_field in fields(self):
            if dc_field.init:
                field_name = dc_field.name
                to_join.append(
                    f'{indentation}{field_name}: {getattr(self, field_name)}')

        if self.id_ in _encountered_ids:
            to_join.append(
                f'{indentation}... Recursion detected; omitting subtrees.')

        else:
            for route in self:
                to_join.append(
                    f'{indentation}++  {route[0:2]}')
                to_join.append(route[2].stringify(
                    depth=depth + 1,
                    _encountered_ids=_encountered_ids | {self.id_}))

        return '\n'.join(to_join)

    def is_equivalent(
            self,
            other: SlotTreeNode,
            *,
            _previous_encounters: dict[int, int] | None = None
            ) -> bool:
        """This compares two slot tree nodes recursively, ignoring IDs.
        If they are otherwise identical -- including the structure of
        any recursive loops -- returns True.

        Note that this is optimized for maintainability and not
        performance. Its primary intended use is in testing.
        """
        previous_encounters: dict[int, int]
        if _previous_encounters is None:
            previous_encounters = {}
        else:
            # Note that the copy here is important because otherwise we'd be
            # mutating state even in other tree branches which might not have
            # encountered us (since we reuse this in the recursive case)
            previous_encounters = {**_previous_encounters}

        # These two cases are trivial: first, if the types aren't the same,
        # they're not equivalent, period. Second, if we already encountered
        # that node's ID, that means we're in a recursive loop, and the two
        # IDs must match.
        if type(self) is not type(other):
            return False
        if other.id_ in previous_encounters:
            return previous_encounters[other.id_] == self.id_

        # First check all the fields, since this should in theory be quick.
        # Yes we could do a bool for this, but doing it as a dict makes it
        # easier to add in temporary print debugs if required
        fields_match: dict[str, bool] = {}
        for dc_field in fields(self):
            field_name = dc_field.name
            if dc_field.init and field_name != 'id_':
                fields_match[field_name] = (
                    getattr(self, field_name) == getattr(other, field_name))
        if not all(fields_match.values()):
            return False

        # Now make sure the routes are the same. This lets us simplify the
        # recursive comparison logic; we don't need any checks to make sure
        # that there weren't any leftover routes on either self or other.
        # It also lets us short-circuit without recursion if they don't match,
        # but that's not the primary motivation behind it.
        self_routes = set(self._index_by_slot_path)
        other_routes = set(other._index_by_slot_path)
        if self_routes != other_routes:
            return False

        # Now finally we're getting to recursive territory.
        previous_encounters[other.id_] = self.id_
        for slot_path in self._index_by_slot_path:
            self_index = self._index_by_slot_path[slot_path]
            self_subtree = self[self_index].subtree
            other_index = other._index_by_slot_path[slot_path]
            other_subtree = other[other_index].subtree

            # Recursion is glorious as long as you don't need to worry about
            # performance! Too bad we can't use it during rendering (because
            # it's way too slow)
            if not self_subtree.is_equivalent(
                other_subtree,
                _previous_encounters=previous_encounters
            ):
                return False

        return True

    def extend(self, routes: Iterable[SlotTreeRoute[T]]):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __delitem__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __setitem__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def clear(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def copy(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def insert(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def pop(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def remove(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def reverse(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __imul__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')

    def __iadd__(self, *args, **kwargs):
        raise ZeroDivisionError('Templatey internal error: not implemented!')


type ConcreteSlotTreeNode = SlotTreeNode[ConcreteSlotTreeNode]


@dataclass(kw_only=True, slots=True)
class PendingSlotTreeNode(SlotTreeNode['PendingSlotTreeNode']):
    # Note: the str is the slot_name that the route needs to be inserted under
    insertion_slot_names: set[str] = field(default_factory=set)

    @property
    def is_insertion_point(self) -> bool:
        return bool(self.insertion_slot_names)


@dataclass(kw_only=True, slots=True)
class DynamicClassSlotTreeNode(SlotTreeNode['DynamicClassSlotTreeNode']):
    # Note: the str is the slot_name that the route needs to be inserted under
    dynamic_class_slot_names: set[str] = field(default_factory=set)

    @property
    def has_dynamic_class_slots(self) -> bool:
        return bool(self.dynamic_class_slot_names)


@dataclass(frozen=True, slots=True)
class PendingSlotTreeContainer:
    """Remember that the point here is to eventually build a lookup from
    a ``{type[template]: SlotTreeNode}``. And we're dealing with
    forward references to the ``type[template]``, meaning we don't have
    a key to use for the slot tree lookup. End of story.

    So what we're doing here instead, is constructing the slot tree as
    best as we can, and keeping track of what nodes need to be populated
    by the forward reference, once it is resolved.

    When the forward ref is resolved, we can simple copy the tree into
    all of the insertion nodes, and then store the pending slot tree
    in the slot tree lookup using the resolved template class.
    """
    pending_slot_type: ForwardRefLookupKey
    pending_root_node: PendingSlotTreeNode


@overload
def _copy_slot_tree[T: SlotTreeNode](
        src_tree: T,
        ) -> T: ...
@overload
def _copy_slot_tree[T: SlotTreeNode](
        src_tree: SlotTreeNode,
        *,
        with_node_type: type[T],
        ) -> T: ...
def _copy_slot_tree[ST: SlotTreeNode, NT: SlotTreeNode](
        src_tree: ST,
        *,
        with_node_type: type[NT] | None = None
        ) -> ST | NT:
    """This creates a copy of an existing slot tree. We use it when
    merging nested slot trees into enclosers; otherwise, we end up with
    a huge mess of "it's not clear what object holds which slot tree"
    that is very difficult to reason about. This is slightly more memory
    intensive, but... again, this is much, much easier to reason about.

    Take special note that this preserves reference cycles, which is a
    bit of a tricky thing.

    Note: if ``into_tree`` is provided, this copies inplace and returns
    the ``into_tree``. Otherwise, a new tree is created and returned.
    In both cases, we also return a lookup from
    ``{old_node.id_: copied_node}``.
    """
    copied_tree: ST | NT
    if with_node_type is None:
        node_type = type(src_tree)
    else:
        node_type = with_node_type

    copied_tree = node_type()
    copied_tree.merge_fields_only(src_tree)

    # This converts ``old_node.id_`` to the new node instance; it's how we
    # implement copying reference cycles
    transmogrified_nodes: dict[int, ST | NT] = {src_tree.id_: copied_tree}
    copy_stack: \
        list[_SlotTreeTraversalFrame[ST | NT, ST]] = [
            _SlotTreeTraversalFrame(
                next_subtree_index=0,
                existing_subtree=copied_tree,
                insertion_subtree=src_tree,
                first_encounters={})]

    while copy_stack:
        current_stack_frame = copy_stack[-1]
        if current_stack_frame.exhausted:
            copy_stack.pop()
            continue

        next_slot_route = current_stack_frame.insertion_subtree[
            current_stack_frame.next_subtree_index]
        next_slot_name, next_slot_type, next_subtree = next_slot_route
        # Do this ASAP so that we don't accidentally forget it somehow
        current_stack_frame.next_subtree_index += 1

        next_subtree_id = next_subtree.id_
        already_copied_node = transmogrified_nodes.get(next_subtree_id)
        # This could be either the first time we hit a recursive subtree,
        # or a non-recursive subtree.
        if already_copied_node is None:
            new_subtree = node_type()
            new_subtree.merge_fields_only(next_subtree)

            if next_subtree.requires_transmogrification:
                transmogrified_nodes[next_subtree_id] = new_subtree

            current_stack_frame.existing_subtree.append(
                SlotTreeRoute.new(
                    next_slot_name,
                    next_slot_type,
                    new_subtree,))
            copy_stack.append(_SlotTreeTraversalFrame(
                next_subtree_index=0,
                existing_subtree=new_subtree,
                insertion_subtree=next_subtree,
                # Note that, though this isn't correct, it also
                # doesn't matter for purely COPYING a slot tree.
                first_encounters={}))

        # We've hit a recursive subtree -- one that we've already copied --
        # which means we don't need to copy it again; instead, we just need to
        # transmogrify the reference so that the nested route refers back to
        # the original copied node.
        else:
            current_stack_frame.existing_subtree.append(
                SlotTreeRoute.new(
                    next_slot_name,
                    next_slot_type,
                    already_copied_node,))

    return copied_tree


def merge_into_slot_tree[T: SlotTreeNode](
        existing_tree_template_cls: TemplateClass,
        # Note: this is only used for determining ``is_terminus``.
        # Also note: None for pending slots. These can never be a terminus
        # anyways, so it doesn't matter.
        existing_tree_slot_type: TemplateClass | None,
        existing_tree: T,
        to_merge: T
        ) -> None:
    """This traverses the existing tree, merging in the slot_name and
    its subtrees into the correct locations in the existing slot tree,
    recursively.

    This will cull any recursion loops to their minimum possible size.
    Note that this can result in different slot tree recursion loop
    "phases", based on which of the two slot types is encountered first.

    Also note that this merges all public slot tree node attributes,
    with three exceptions:
    ++  the ``id_`` of the existing tree is always preserved
    ++  ``is_recursive`` will always be calculated based on the
        recursion loop culling
    ++  ``is_terminus`` will always be calculated based on node types
        matching the passed ``existing_tree_slot_type``.
    """
    node_type = type(existing_tree)
    transmogrified_nodes: dict[int, T] = {to_merge.id_: existing_tree}
    # Note: we want our internal culling logic to handle ``is_recursive``.
    # Also note: the existing tree usually knows whether or not it's already
    # recursive, but there's an edge case where it doesn't:
    # 1.. have a recursive loop between two templates
    # 2.. when resolving the back-ref of the second template, it finds a
    #     forward ref to itself
    # 3.. it now immediately resolves the forward ref to itself
    # In this situation, the short-circuited self-referential forward ref
    # results in is_terminus being False, which we correct here.
    existing_tree.merge_fields_only(
        to_merge, fields_to_skip={'is_recursive', 'is_terminus'})
    existing_tree.is_terminus = (
        existing_tree_template_cls is existing_tree_slot_type)

    # Counterintuitive: since we're MERGING trees, the existing_subtree is
    # actually the DESTINATION, and the insertion_subtree the source!
    merge_stack: list[_SlotTreeTraversalFrame[T, T]] = [
        _SlotTreeTraversalFrame(
            next_subtree_index=0,
            existing_subtree=existing_tree,
            insertion_subtree=to_merge,
            first_encounters={existing_tree_template_cls: existing_tree})]
    # Yes, in theory, this one specific operation of merging trees would be
    # faster if the trees were dicts instead of iterative structures. But
    # we're not optimizing for tree merging; we're optimizing for rendering!
    # And in that case, we're better off with a simple iterative structure.
    while merge_stack:
        current_stack_frame = merge_stack[-1]
        if current_stack_frame.exhausted:
            merge_stack.pop()
            continue

        existing_subtree = current_stack_frame.existing_subtree
        next_slot_route = current_stack_frame.insertion_subtree[
            current_stack_frame.next_subtree_index]
        next_slot_name, next_slot_type, next_subtree = next_slot_route
        # Do this ASAP so that we don't accidentally forget it somehow (also
        # because we want to use a continue statement in a second)
        current_stack_frame.next_subtree_index += 1

        next_subtree_id = next_subtree.id_
        already_merged_node = transmogrified_nodes.get(next_subtree_id)

        # For merging, we're going to handle the recursive subtree case first,
        # because it makes the rest of the logic cleaner
        if already_merged_node is not None:
            # Note that appending is idempotent as long as the
            # already_merged_node is the same object as any hypothetical
            # already_existing_already_merged_node that could be found there.
            # So this will only error if the slot path already exists, AND
            # it points to a different node.
            current_stack_frame.existing_subtree.append(
                SlotTreeRoute.new(
                    next_slot_name, next_slot_type, already_merged_node))

            already_merged_node.merge_fields_only(
                next_subtree,
                # Note: we want our internal culling logic to handle
                # ``is_recursive``. Otherwise, we might result in something
                # being reported as recursive, which actually isn't -- because
                # the ... phase, I guess you could say ... of the recursion
                # loop is different with a different root node.
                # Also note: we don't want to preserve the terminus value of
                # a tree we're merging in; we want the terminus to be purely
                # calculated within the merge function. This prevents bugs
                # when merging trees across slot types, which happens a lot
                # with pending slot trees.
                fields_to_skip={'is_recursive', 'is_terminus'})
            # This might seem redundant, but there are some weird edge cases
            # when merging a fully-defined tree (for example, adding in
            # self-referential recursion loops to a pending tree) that require
            # it to be set.
            # (This is because we skip it when merging the fields on the
            # existing tree; this recovers the correct state).
            already_merged_node.is_recursive = True
            continue

        # This accomplishes two things: first, it culls an extra cycle from the
        # to_merge tree that would ultimately have the same effect. Secondly,
        # it ensures correct recursion when we're merging in the pending tree
        # from a class that used us as a forward reference, since the other
        # class won't yet be resolved.
        # Note that, by definition, this won't happen the first time we
        # encounter a node of any particular type, including the type of the
        # rootwards-most node for the whole tree. This is expected! We first
        # need to encounter the slot type before we can recurse to it. **We
        # will never recurse back to the root of the slot tree!**
        if next_slot_type in current_stack_frame.first_encounters:
            # Note that we want to avoid modifying the already-encountered
            # tree. This is mostly because we copy trees willy nilly and want
            # to avoid them accidentally getting out of sync. However, it does
            # require us to be very careful to ensure that we've corrently
            # copied all needed values over prior to merging trees, if we've
            # just created the existing tree from scratch.
            next_existing_subtree = current_stack_frame.first_encounters[
                next_slot_type]
            next_existing_subtree.is_recursive = True

            next_existing_route = SlotTreeRoute.new(
                next_slot_name,
                next_slot_type,
                next_existing_subtree)
            if existing_subtree.has_route_for(next_slot_name, next_slot_type):
                existing_subtree.rewrite_route_for(
                    next_slot_name, next_slot_type, next_existing_route)
                # raise RecursionError(
                #     'Non-culled infinite recursion while merging templatey '
                #     + 'slot trees!', next_slot_name, next_slot_type)

            else:
                existing_subtree.append(next_existing_route)

            # Note that we don't need to update transmogrification for two
            # reasons: first, because we added the root node at the very
            # beginning, and second, because we -- by definition -- cannot
            # have any deeper references to this part of the destination tree,
            # because  we're culling the depthwise-rest of the source tree
            continue

        # The existing subtree -- the one we're merging INTO -- has a route for
        # this slot name and type already, so we need to merge them together
        # instead of simply copy/transmogrify/cull
        elif existing_subtree.has_route_for(next_slot_name, next_slot_type):
            next_existing_route = existing_subtree.get_route_for(
                    next_slot_name, next_slot_type)
            __, __, next_existing_subtree = next_existing_route
            next_existing_subtree.merge_fields_only(
                next_subtree,
                # Note: we want our internal culling logic to handle
                # ``is_recursive``. Otherwise, we might result in something
                # being reported as recursive, which actually isn't -- because
                # the ... phase, I guess you could say ... of the recursion
                # loop is different with a different root node.
                # Also note: we don't want to preserve the terminus value of
                # a tree we're merging in; we want the terminus to be purely
                # calculated within the merge function. This prevents bugs
                # when merging trees across slot types, which happens a lot
                # with pending slot trees.
                fields_to_skip={'is_recursive', 'is_terminus'})

        # The existing subtree doesn't have any existing routes for this, so
        # we don't need to worry about merging things together -- but we still
        # need to worry about transmogrification and culling.
        # Also note that there might be identically-named slots for different
        # slot types in the case of a union, but that will be handled on a
        # different iteration of the merge stack while loop.
        else:
            # Note that the next_subtree might have a different node type
            # than the existing tree!
            next_existing_subtree = node_type()
            next_existing_subtree.merge_fields_only(
                next_subtree,
                fields_to_skip={'is_recursive', 'is_terminus'})
            # Because we do lots and lots of merging, with offsets and
            # transforms etc, it's easiest to just reset this every time.
            # This helps prevent weird bugs with copying.
            next_existing_subtree.is_terminus = (
                # Note: what matters here is the slot type, not the template
                # class!
                next_slot_type is existing_tree_slot_type)
            next_existing_route = SlotTreeRoute.new(
                next_slot_name,
                next_slot_type,
                next_existing_subtree)
            existing_subtree.append(next_existing_route)

        if next_existing_subtree.requires_transmogrification:
            transmogrified_nodes[next_subtree_id] = next_existing_subtree

        if next_subtree:
            next_first_encounters = {**current_stack_frame.first_encounters}
            next_first_encounters.setdefault(
                next_slot_type, next_existing_subtree)
            merge_stack.append(_SlotTreeTraversalFrame(
                next_subtree_index=0,
                existing_subtree=next_existing_subtree,
                insertion_subtree=next_subtree,
                first_encounters=next_first_encounters))


def update_encloser_with_trees_from_slot(
        enclosing_cls: TemplateClass,
        enclosing_slot_tree_lookup: dict[TemplateClass, ConcreteSlotTreeNode],
        enclosing_pending_ref_lookup:
            dict[ForwardRefLookupKey, PendingSlotTreeContainer],
        enclosing_cls_ref_lookup_key: ForwardRefLookupKey,
        nested_slot_type: TemplateClass,
        nested_slot_offset: PendingSlotTreeNode,
        ) -> None:
    """This function updates the enclosing class' concrete and pending
    slot trees with the concrete and pending slot trees from the
    nested slot.

    Note that it can be called during initial signature assembly, hence
    needing explicit arguments instead of relying upon the lookups
    already being defined on the class.
    """
    nested_slot_xable = cast(type[TemplateIntersectable], nested_slot_type)
    nested_lookup = nested_slot_xable._templatey_signature._slot_tree_lookup
    nested_pending_refs = (
        nested_slot_xable._templatey_signature._pending_ref_lookup)
    # First and foremost, we can't forget to add a route for the
    # actual slot itself!
    _apply_concrete_insertions(
        enclosing_cls,
        enclosing_slot_tree_lookup,
        nested_slot_type,
        nested_lookup,
        nested_slot_type,
        # Note that this is actually supposed to be the nested slot offset
        # in both cases; the function itself gets the appropriate tree from
        # the nested_lookup
        nested_slot_offset)

    # Now the CONCRETE nested slots.
    # IMPORTANT: all of the nested concrete slots need to already
    # be merged in before starting the nested pending refs! (More
    # info below).
    for doubly_nested_slot_type in nested_lookup:
        _apply_concrete_insertions(
            enclosing_cls,
            enclosing_slot_tree_lookup,
            nested_slot_type,
            nested_lookup,
            doubly_nested_slot_type,
            # Note that this is actually supposed to be the nested slot offset
            # in both cases; the function itself gets the appropriate tree from
            # the nested_lookup
            nested_slot_offset)

    # And now, finally, the nested PENDING slots.
    # IMPORTANT: all of the nested concrete slots need to already
    # be merged in before starting the nested pending refs!
    # If we reversed the order, then we wouldn't be able to
    # guarantee that we'd already discovered every possible
    # terminus node for the nested slot type (due to potential
    # recursion loops). This could then cause us to miss out on
    # a lot of the combinatorics of the offset pending trees.
    # That being said, note that this is just a transform-and-merge
    # operation; we're not adding any additional insertion points
    # on top of what the nested slot defines. That happens later,
    # when we deal with the pending ref defs for the enclosing
    # template class (the one we're currently constructing a
    # signature for).

    # Note: we want to special-case a nested forward ref to the
    # enclosing class (ie, a recursion loop) for exactly the
    # same reason: to guarantee it's resolved before we start on
    # the other nested ones. Hence copy-and-mutate.
    unresolved_nested_forward_refs = {**nested_pending_refs}

    # Remember that we're potentially in the middle of constructing the
    # signature for a new template class. If the nested class
    # (from the slot) was depending on the class we're still
    # constructing, it hasn't yet been updated with the
    # resolved class. Therefore, instead of needing to come
    # back and fix up any recursive forward refs later, we can
    # simply do them right here, right now.
    # Also note that we'll NEVER have an existing pending tree
    # for this, because we're adding it directly and
    # immediately into the actual slot tree.
    encloser_referencing_pending_container = (
        unresolved_nested_forward_refs.pop(
            enclosing_cls_ref_lookup_key, None))
    if encloser_referencing_pending_container is not None:
        # This is effectively a special-case of pending slot tree extension
        # where we immediately apply the concrete insertions. But we still need
        # to merge the two pending trees (which handles the combinatorics --
        # remember that each insertion point for the nested class will also
        # get N insertion points from its references to the enclosing cls).
        _extend_pending_slot_tree(
            enclosing_cls=enclosing_cls,
            enclosing_slot_tree_lookup=enclosing_slot_tree_lookup,
            enclosing_pending_ref_lookup=enclosing_pending_ref_lookup,
            nested_cls=nested_slot_type,
            nested_pending_container=encloser_referencing_pending_container,
            nested_pending_ref=enclosing_cls_ref_lookup_key,)
        offset_pending_container = enclosing_pending_ref_lookup.pop(
            enclosing_cls_ref_lookup_key)
        _apply_concrete_insertions(
            enclosing_cls,
            enclosing_slot_tree_lookup,
            # These are counter-intuitive, but correct. We've divorced this
            # from the nested slot entirely; all the insertion points for it
            # now are pointed back at ourselves.
            enclosing_cls,
            enclosing_slot_tree_lookup,
            enclosing_cls,
            offset_pending_container.pending_root_node)

    # Okay, after all of that meticulous work, we can now guarantee
    # that all of the concrete classes we know about up until this
    # point have been fully resolved and incorporated into the
    # slot tree. That means we can FINALLY start merging the
    # pending items from the nested slot.
    for (
        nested_forward_ref_key, nested_pending_container
    ) in unresolved_nested_forward_refs.items():
        # Okay, so: technically, we should be checking the concrete slot tree
        # for any reference loops to the enclosing class before doing this.
        # The problem is that this can change literally every time a pending
        # ref is resolved. Therefore, we do that THEN -- and quite literally.
        # Every time a pending slot is resolved, after resolution is complete,
        # we go through every tree and merge in the recursive one.
        _extend_pending_slot_tree(
            enclosing_cls=enclosing_cls,
            enclosing_slot_tree_lookup=enclosing_slot_tree_lookup,
            enclosing_pending_ref_lookup=enclosing_pending_ref_lookup,
            nested_cls=nested_slot_type,
            nested_pending_container=nested_pending_container,
            nested_pending_ref=nested_forward_ref_key,)


def _apply_concrete_insertions(
        enclosing_cls: TemplateClass,
        enclosing_slot_tree_lookup: dict[TemplateClass, ConcreteSlotTreeNode],
        nested_slot_type: TemplateClass,
        # Note: may or may not be different from insertion_cls; see note in
        # docstring
        nested_slot_tree_lookup: dict[TemplateClass, ConcreteSlotTreeNode],
        insertion_slot_type: TemplateClass,
        insertion_tree: PendingSlotTreeNode
        ) -> None:
    """This takes an insertion tree (a pending slot tree node), applies
    a resolved, concrete insertion class at each of the insertion
    points on the insertion tree, and then merges the result into the
    slot tree lookup for the enclosing class.

    Note that this is meant to be used both with the nested class
    itself, as well as its (doubly-)nested concrete slots. In the former
    case, the slot tree lookup will match the insertion class. In the
    latter case, the insertion class will instead be the (doubly-)nested
    concrete slot.
    """
    root_after_insertion = SlotTreeNode()
    tree_to_insert = nested_slot_tree_lookup.get(
        insertion_slot_type,
        SlotTreeNode(is_terminus=True))

    # This converts ``old_node.id_`` to the new node instance; it's how we
    # implement copying reference cycles
    transmogrified_nodes: dict[int, SlotTreeNode] = {
        insertion_tree.id_: root_after_insertion}
    stack: \
        list[_SlotTreeTraversalFrame[SlotTreeNode, PendingSlotTreeNode]] = [
        _SlotTreeTraversalFrame(
            next_subtree_index=0,
            existing_subtree=root_after_insertion,
            insertion_subtree=insertion_tree,
            first_encounters={enclosing_cls: root_after_insertion})]

    while stack:
        current_stack_frame = stack[-1]
        src_subtree = current_stack_frame.insertion_subtree
        target_subtree = current_stack_frame.existing_subtree

        if current_stack_frame.exhausted:
            # This is counter-intuitive. Note that:
            # ++  the stack is for checking nested (deeper) routes only. it
            #     does not check for insertions on the current stack frame.
            # ++  the current stack frame might, though, actually have some
            #     insertions!
            # ++  we only want those insertions to be applied ONCE, and not
            #     once per subtree
            # ++  ordering of the insertion application doesn't matter
            # Therefore, we apply insertions to the current frame immediately
            # before discarding the frame for being exhausted.
            for nested_slot_type_slot_name in src_subtree.insertion_slot_names:
                target_subtree.append(
                    SlotTreeRoute.new(
                        nested_slot_type_slot_name,
                        nested_slot_type,
                        _copy_slot_tree(tree_to_insert)))

            stack.pop()
            continue

        next_slot_route = src_subtree[current_stack_frame.next_subtree_index]
        next_slot_name, next_slot_type, next_subtree = next_slot_route
        # Do this ASAP so that we don't accidentally forget it somehow
        current_stack_frame.next_subtree_index += 1

        # Note that this will still get merged into the actual full slot tree
        # for the enclosing template, which will cull any extra links in
        # recursive reference cycles, so we don't need to worry about that
        # here.
        next_subtree_id = next_subtree.id_
        transmogrified_dest_subtree = transmogrified_nodes.get(next_subtree_id)

        # This could be either the first time we hit a recursive subtree,
        # or a non-recursive subtree.
        if transmogrified_dest_subtree is None:
            dest_subtree = SlotTreeNode()
            dest_subtree.merge_fields_only(next_subtree)

            if next_subtree.requires_transmogrification:
                transmogrified_nodes[next_subtree_id] = dest_subtree

            # Note that, since we're building a new tree from scratch, we don't
            # need to worry about this already existing.
            target_subtree.append(
                SlotTreeRoute.new(
                    next_slot_name,
                    next_slot_type,
                    dest_subtree))
            next_first_encounters = {**current_stack_frame.first_encounters}
            next_first_encounters.setdefault(next_slot_type, dest_subtree)
            stack.append(_SlotTreeTraversalFrame(
                next_subtree_index=0,
                existing_subtree=dest_subtree,
                insertion_subtree=next_subtree,
                first_encounters=next_first_encounters))

        # We've hit a recursive subtree -- one that we've already copied --
        # which means we don't need to copy it again; instead, we just need to
        # transmogrify the reference so that the nested route refers back to
        # the original copied node.
        else:
            dest_subtree = transmogrified_dest_subtree
            # Note: is_recursive was already set!
            target_subtree.append(
                SlotTreeRoute.new(
                    next_slot_name,
                    next_slot_type,
                    transmogrified_dest_subtree))

    enclosing_root = enclosing_slot_tree_lookup.get(insertion_slot_type)
    if enclosing_root is None:
        enclosing_root = enclosing_slot_tree_lookup[insertion_slot_type] = (
            SlotTreeNode())

    merge_into_slot_tree(
        enclosing_cls,
        insertion_slot_type,
        enclosing_root,
        root_after_insertion)


def _extend_pending_slot_tree(
        enclosing_cls: TemplateClass,
        enclosing_slot_tree_lookup: dict[TemplateClass, ConcreteSlotTreeNode],
        enclosing_pending_ref_lookup:
            dict[ForwardRefLookupKey, PendingSlotTreeContainer],
        nested_cls: TemplateClass,
        nested_pending_container: PendingSlotTreeContainer,
        nested_pending_ref: ForwardRefLookupKey,
        ) -> None:
    """When we resolve a pending class, we do (of course) resolve the
    pending class itself. But in the process, we may also suddenly have
    a bunch of new pending refs, courtesy of the pending class.

    At first glance, you might think that we could simply dump these
    into the enclosing pending ref lookup, offset by the slot path for
    the newly-resolved class. HOWEVER, this ignores recursion, and is
    fragile in situations where classes have multiple slot names for
    a particular slot type.

    Instead, what we really need to do, is:
    ++  create a pending tree copy of the concrete slot tree for the
        newly-resolved class
    ++  at every terminus:
        ++  set ``is_terminus`` to False
        ++  insert a copy of the nested root tree
    """
    src_concrete_tree = enclosing_slot_tree_lookup[nested_cls]
    dest_copied_tree = _copy_slot_tree(
        src_concrete_tree,
        with_node_type=PendingSlotTreeNode)
    # Note that it's not an issue that this doesn't track the stack, because
    # we aren't changing the structure of the existing tree at all.
    # (We can't reuse first_encounters because it's meant for a TemplateClass)
    encountered_node_ids: set[int] = set()

    stack: \
        list[
            _SlotTreeTraversalFrame[
                PendingSlotTreeNode, ConcreteSlotTreeNode]] = [
        _SlotTreeTraversalFrame(
            next_subtree_index=0,
            existing_subtree=dest_copied_tree,
            insertion_subtree=src_concrete_tree,
            first_encounters={})]

    while stack:
        current_stack_frame = stack[-1]
        if current_stack_frame.exhausted:
            stack.pop()
            continue

        next_slot_route = current_stack_frame.insertion_subtree[
            current_stack_frame.next_subtree_index]
        next_slot_name, next_slot_type, next_subtree = next_slot_route
        # Note that we can't rely upon the indices being the same, because as
        # soon as we make an insertion, they'll drift out of sync
        target_subtree = current_stack_frame.existing_subtree.get_route_for(
            next_slot_name, next_slot_type).subtree

        # Do this ASAP so that we don't accidentally forget it somehow
        current_stack_frame.next_subtree_index += 1

        # There's never anything to do here; it means we've already done all
        # of our transforms and insertions, because it's always recursive.
        # Doesn't matter where on the tree we are. We've already fixed this
        # node, period, end of story.
        # (Don't forget that the weird doubling-back we do only applies to the
        # next EXISTING subtree, but still advances the next (insertion)
        # subtree. So we're safe in that regard.)
        if next_subtree.id_ in encountered_node_ids:
            continue
        encountered_node_ids.add(next_subtree.id_)

        if next_subtree.is_terminus:
            # As per docstring, the idea here is that we're converting every
            # terminus into an insertion point for a different (pending) slot.
            # Therefore it's no longer a terminus, since the slot is different.
            target_subtree.is_terminus = False
            merge_into_slot_tree(
                # Note: this is intentionally NOT the enclosing_cls! We're
                # working on an offset tree here, and we need to make sure that
                # we correctly report the class of the OFFSET root!
                next_slot_type,
                None,
                target_subtree,
                nested_pending_container.pending_root_node)

        # Whether or not we found a terminus, we need to check all of the
        # possibilities on the next subtree. (As a reminder, you can have a
        # terminus that still has nested nodes!)
        stack.append(_SlotTreeTraversalFrame(
            next_subtree_index=0,
            existing_subtree=target_subtree,
            insertion_subtree=next_subtree,
            # Allow the final merging to handle any culling; we don't need to
            # do it here.
            first_encounters={}))

    # Okay, the transform is complete; now we just need to insert it into
    # the actual pending slot tree for the enclosing template class.
    if nested_pending_ref in enclosing_pending_ref_lookup:
        dest_pending_tree_container = enclosing_pending_ref_lookup[
            nested_pending_ref]
    else:
        dest_pending_tree_container = enclosing_pending_ref_lookup[
            nested_pending_ref] = PendingSlotTreeContainer(
                pending_slot_type=nested_pending_ref,
                pending_root_node=PendingSlotTreeNode())

    merge_into_slot_tree(
        enclosing_cls,
        None,
        dest_pending_tree_container.pending_root_node,
        dest_copied_tree)


@dataclass(slots=True)
class _SlotTreeTraversalFrame[ET: SlotTreeNode, IT: SlotTreeNode]:
    next_subtree_index: int
    existing_subtree: ET
    insertion_subtree: IT
    first_encounters: dict[TemplateClass | None, ET]

    _insertion_subtree_len: int = field(init=False, repr=False, compare=False)

    def __post_init__(self):
        self._insertion_subtree_len = len(self.insertion_subtree)

    @property
    def exhausted(self) -> bool:
        """Returns True if the to_merge_subtree has been exhausted, and
        there are no more subtrees to merge.
        """
        return self.next_subtree_index >= self._insertion_subtree_len


def gather_dynamic_class_slots(
        enclosing_template_cls: TemplateClass,
        enclosing_dynamic_slot_names: set[str],
        enclosing_slot_tree_lookup: dict[TemplateClass, ConcreteSlotTreeNode]
        ) -> DynamicClassSlotTreeNode:
    """This is responsible for combining all of the dynamic slots on
    the enclosing template class with all of the dynamic slots on any
    nested templates (appropriately offset from the enclosing class,
    of course).

    This will actually create the slot tree node; if you need to merge
    into an existing one (because you've resolved a forward ref), see
    ``merge_dynamic_slots``.

    Note that this can only be called **after** the enclosing slot tree
    lookup has been fully populated.
    """
    tree_root = DynamicClassSlotTreeNode(
        dynamic_class_slot_names=enclosing_dynamic_slot_names)
    # This may or may not be present in the lookup. If it is, we need to
    # handle it like any other slot, because there might be recursion loops.
    # If not... well, then we'll just skip it in the for loop.
    # (This seeming redundancy was the most practical way to do this).
    if (
        enclosing_template_cls in enclosing_slot_tree_lookup
        and enclosing_dynamic_slot_names
    ):
        merge_dynamic_class_slots(
            enclosing_template_cls,
            enclosing_slot_tree_lookup,
            tree_root,
            enclosing_template_cls,
            tree_root)

    for nested_template_cls in enclosing_slot_tree_lookup:
        # Note: more than just deduplication; it literally won't be available
        # at this point, because the signature won't be defined yet
        if nested_template_cls is not enclosing_template_cls:
            nested_template_signature = cast(
                type[TemplateIntersectable],
                nested_template_cls)._templatey_signature
            # Only merge if it actually HAS dynamic class slots! Otherwise,
            # we'd be wasting a bunch of tree traversals during render time!
            if nested_template_signature.dynamic_class_slot_names:
                merge_dynamic_class_slots(
                    enclosing_template_cls,
                    enclosing_slot_tree_lookup,
                    tree_root,
                    nested_template_cls,
                    nested_template_signature._dynamic_class_slot_tree)

    return tree_root


def merge_dynamic_class_slots(
        enclosing_template_cls: TemplateClass,
        enclosing_slot_tree_lookup: dict[TemplateClass, ConcreteSlotTreeNode],
        enclosing_dynamic_class_slot_tree: DynamicClassSlotTreeNode,
        nested_template_cls: TemplateClass,
        # This needs to be explicit, because it's not always available on the
        # nested_template_cls (yet -- if enclosing_template_cls is
        # nested_template_cls, we might still be in initial signature creation)
        nested_dynamic_class_slot_tree: DynamicClassSlotTreeNode
        ) -> None:
    """After the slot tree of a newly-resolved nested template class has
    been fully merged, call this to also update the enclosing dynamic
    slots slot tree for that particular nested_template_cls.

    Note that this modifies the tree in-place!

    This is very similar to _extend_pending_slot_tree. The general idea
    is that we search for every terminus of the existing concrete tree
    for the nested template class, and then at that terminus, we set
    the terminus to False (since it isn't used by dynamic class slots)
    and insert a copy of the nested template class' dynamic class slot
    tree.
    """
    src_concrete_tree = enclosing_slot_tree_lookup[nested_template_cls]
    dest_copied_tree = _copy_slot_tree(
        src_concrete_tree,
        with_node_type=DynamicClassSlotTreeNode)
    # Note that it's not an issue that this doesn't track the stack, because
    # we aren't changing the structure of the existing tree at all.
    # (We can't reuse first_encounters because it's meant for a TemplateClass)
    encountered_node_ids: set[int] = set()

    stack: \
        list[
            _SlotTreeTraversalFrame[
                DynamicClassSlotTreeNode, ConcreteSlotTreeNode]] = [
        _SlotTreeTraversalFrame(
            next_subtree_index=0,
            existing_subtree=dest_copied_tree,
            insertion_subtree=src_concrete_tree,
            first_encounters={})]

    while stack:
        current_stack_frame = stack[-1]
        if current_stack_frame.exhausted:
            stack.pop()
            continue

        next_slot_route = current_stack_frame.insertion_subtree[
            current_stack_frame.next_subtree_index]
        next_slot_name, next_slot_type, next_subtree = next_slot_route
        # Note that we can't rely upon the indices being the same, because as
        # soon as we make an insertion, they'll drift out of sync
        target_subtree = current_stack_frame.existing_subtree.get_route_for(
            next_slot_name, next_slot_type).subtree

        # Do this ASAP so that we don't accidentally forget it somehow
        current_stack_frame.next_subtree_index += 1

        # There's never anything to do here; it means we've already done all
        # of our transforms and insertions, because it's always recursive.
        # Doesn't matter where on the tree we are. We've already fixed this
        # node, period, end of story.
        # (Don't forget that the weird doubling-back we do only applies to the
        # next EXISTING subtree, but still advances the next (insertion)
        # subtree. So we're safe in that regard.)
        if next_subtree.id_ in encountered_node_ids:
            continue
        encountered_node_ids.add(next_subtree.id_)

        if next_subtree.is_terminus:
            # As per docstring, the idea here is that we're converting every
            # terminus into an insertion point for a different (pending) slot.
            # Therefore it's no longer a terminus, since the slot is different.
            target_subtree.is_terminus = False
            merge_into_slot_tree(
                # Note: this is intentionally NOT the enclosing_cls! We're
                # working on an offset tree here, and we need to make sure that
                # we correctly report the class of the OFFSET root!
                next_slot_type,
                None,
                target_subtree,
                nested_dynamic_class_slot_tree)

        # Whether or not we found a terminus, we need to check all of the
        # possibilities on the next subtree. (As a reminder, you can have a
        # terminus that still has nested nodes!)
        stack.append(_SlotTreeTraversalFrame(
            next_subtree_index=0,
            existing_subtree=target_subtree,
            insertion_subtree=next_subtree,
            # Allow the final merging to handle any culling; we don't need to
            # do it here.
            first_encounters={}))

    merge_into_slot_tree(
        enclosing_template_cls,
        None,
        enclosing_dynamic_class_slot_tree,
        dest_copied_tree)


@dataclass(slots=True)
class _DynaClsExtractorFrame:
    active_instance: TemplateParamsInstance
    active_subtree: DynamicClassSlotTreeNode
    target_subtree_index: int
    target_instance_index: int
    target_instances_count: int = field(kw_only=True, default=0)
    target_instances: Sequence[TemplateParamsInstance] = field(
        kw_only=True, init=False)

    # See note in extract_dynamic_class_slot_types for why this is necessary
    # in addition to the encountered instance IDs
    direct_recursion_guard: set[int]
    target_slot_name_index: int = 0

    _active_subtree_len: int = field(init=False, repr=False, compare=False)
    _ordered_slot_names: list[str] = field(
        init=False, repr=False, compare=False)
    _ordered_slot_names_len: int = field(init=False, repr=False, compare=False)

    def __post_init__(self):
        self._active_subtree_len = len(self.active_subtree)
        self._ordered_slot_names = ordered_slot_names = cast(
            TemplateIntersectable, self.active_instance
        )._templatey_signature._ordered_dynamic_class_slot_names
        self._ordered_slot_names_len = len(ordered_slot_names)

    @property
    def subtrees_exhausted(self) -> bool:
        return self.target_subtree_index >= self._active_subtree_len

    @property
    def dynacls_slots_exhausted(self) -> bool:
        return self.target_slot_name_index >= self._ordered_slot_names_len


# Yes, this is really complicated and really long. But unfortunately, the
# combinatorics are really bad, function calls in python are expensive, and
# this is on the critical hot path for rendering. So it's a calculated smell.
def extract_dynamic_class_slot_types(  # noqa: C901, PLR0912, PLR0915
        root_template_instance: TemplateParamsInstance,
        dynamic_class_slot_tree: DynamicClassSlotTreeNode
        ) -> set[TemplateClass]:
    """Given a root template instance and its associated dynamic class
    slot tree, walks the tree and extracts all template classes for
    slots defined as having a dynamic class.

    This does no culling based on whether or not the class was already
    loaded; it simply constructs a set of all encountered dynamic
    template classes.
    """
    encountered_instance_ids: set[int] = set()
    dynamic_template_classes: set[TemplateClass] = set()
    stack: list[_DynaClsExtractorFrame] = [
        _DynaClsExtractorFrame(
            active_instance=root_template_instance,
            active_subtree=dynamic_class_slot_tree,
            target_subtree_index=0,
            target_instance_index=0,
            direct_recursion_guard={id(root_template_instance)})]

    while stack:
        frame = stack[-1]
        active_instance = frame.active_instance
        active_instance_id = id(active_instance)
        active_subtree = frame.active_subtree

        if active_instance_id in encountered_instance_ids:
            stack.pop()
            continue

        # Note: this is also the branch we use for "we only want to do this
        # N times per frame, not N times per subtree"
        if frame.subtrees_exhausted:
            if frame.dynacls_slots_exhausted:
                stack.pop()
                encountered_instance_ids.add(active_instance_id)

            else:
                slot_name = frame._ordered_slot_names[
                    frame.target_slot_name_index]
                target_instance_index = frame.target_instance_index

                # As with subtree checking, this is effectively a nested stack,
                # but we're maintaining state within the current frame to
                # decrease resource usage. See below for more explanation; it's
                # basically the same logic.
                if target_instance_index == 0:
                    target_instances = getattr(active_instance, slot_name)
                    target_instances_count = len(target_instances)

                    if target_instances_count <= 0:
                        frame.target_slot_name_index += 1
                        continue

                    else:
                        frame.target_instances_count = target_instances_count
                        frame.target_instances = target_instances

                else:
                    target_instances_count = frame.target_instances_count
                    if frame.target_instance_index >= target_instances_count:
                        frame.target_instances_count = 0
                        frame.target_instance_index = 0
                        frame.target_slot_name_index += 1
                        continue

                    target_instances = frame.target_instances

                instance_to_check = target_instances[target_instance_index]
                # Note: we need to special-case this separately from indirect
                # recursion, because we can't add the current instance ID to
                # encountered_instance_ids until we're done processing the
                # slots (otherwise we'd skip all of them!).
                # Additionally, this needs to be a stacked-set, and not a
                # simple comparison between instance_to_check and
                # active_instance_id, because otherwise, indirect recursion
                # (A -> B -> A -> ...) will infinitely descend without ever
                # marking anything as encountered. Doing it this way ensures
                # that only the outermost level is responsible for checking A.
                direct_recursion_guard = frame.direct_recursion_guard
                instance_to_check_id = id(instance_to_check)
                if instance_to_check_id not in direct_recursion_guard:
                    xable_to_check = cast(
                        TemplateIntersectable, instance_to_check)
                    dynamic_template_classes.add(type(instance_to_check))
                    # Here we also recurse to check the dynamic-class-slot
                    # instance to check if it, too, has dynamic classes. This
                    # is where the combinatorics really explode, but we're
                    # writing the code this way to keep the size of the stack
                    # minimal.
                    stack.append(_DynaClsExtractorFrame(
                        active_instance=instance_to_check,
                        active_subtree=
                            xable_to_check._templatey_signature
                            ._dynamic_class_slot_tree,
                        target_subtree_index=0,
                        target_instance_index=0,
                        direct_recursion_guard=
                            direct_recursion_guard | {instance_to_check_id}))

                frame.target_instance_index += 1

            continue

        slot_route = active_subtree[frame.target_subtree_index]
        slot_name, slot_type, subtree = slot_route
        target_instance_index = frame.target_instance_index

        # This is, in a way, a nested stack, but we're maintaining
        # the stack state within the _DynaClsExtractorFrame.
        # At any rate, we use the zero-index iteration of the loop to
        # memoize some values on the stack frame.
        if target_instance_index == 0:
            target_instances = getattr(active_instance, slot_name)
            target_instances_count = len(target_instances)

            # Check in advance if there are no target instances at all,
            # and if so, skip the whole thing. This isn't just for
            # performance; the processing logic depends on it.
            if target_instances_count <= 0:
                # Note: this is critical! Otherwise we'll infinitely loop.
                frame.target_subtree_index += 1
                continue
            else:
                frame.target_instances_count = target_instances_count
                frame.target_instances = target_instances

        else:
            target_instances_count = frame.target_instances_count
            # We've exhausted the target instances; reset the state for
            # the next slot tree route and then continue.
            if frame.target_instance_index >= target_instances_count:
                # Note: we're deliberately skipping the target instances
                # themselves, because it'll just get overwritten the next
                # time around, so we can save ourselves an operation.
                frame.target_instances_count = 0
                frame.target_instance_index = 0
                # Note: this is critical! Otherwise we'll infinitely loop.
                frame.target_subtree_index += 1
                continue

            # We still have some instances to target; normalize the state
            # so that we can operate on them.
            target_instances = frame.target_instances

        # Okay, status check: we have our stack frame state configured
        # correctly, and we have target instances to check.
        instance_to_check = target_instances[target_instance_index]
        # Note: exact match here; not subclassing! Subclassing breaks too
        # many things, so we don't support it.
        if type(instance_to_check) is slot_type:
            stack.append(_DynaClsExtractorFrame(
                active_instance=instance_to_check,
                active_subtree=subtree,
                target_subtree_index=0,
                target_instance_index=0,
                direct_recursion_guard=frame.direct_recursion_guard))

        frame.target_instance_index += 1

    return dynamic_template_classes



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import KW_ONLY
from dataclasses import dataclass
from dataclasses import field

from templatey._slot_tree import SlotTreeNode
from templatey._types import TemplateClass
from templatey._types import TemplateInstanceID
from templatey._types import TemplateParamsInstance
from templatey.parser import ParsedTemplateResource
from templatey.parser import TemplateInstanceContentRef
from templatey.parser import TemplateInstanceVariableRef


@dataclass(slots=True, frozen=True)
class ProvenanceNode:
    """ProvenanceNode instances are unique to the exact location
    on the exact render tree of a particular template instance. If the
    template instance gets reused within the same render tree, it will
    have multiple provenance nodes. And each different slot in an
    enclosing template will have a separate provenance node, potentially
    with different namespace overrides.

    This is used both during function execution (to calculate the
    concrete value of any parameters), and also while flattening the
    render tree.

    Also note that the root template, **along with any templates
    injected into the render by environment functions,** will have an
    empty list of provenance nodes.

    Note that, since overrides from the enclosing template can come
    exclusively from the template body -- and are therefore shared
    across all nested children of the same slot -- they don't get stored
    within the provenance, since we'd require access to the template
    bodies, which we don't yet have.
    """
    encloser_slot_key: str
    encloser_slot_index: int
    # The reason to have both the instance and the instance ID is so that we
    # can have hashability of the ID while not imposing an API on the instances
    instance_id: TemplateInstanceID
    instance: TemplateParamsInstance = field(compare=False, repr=False)


@dataclass(slots=True, frozen=True)
class Provenance:
    slotpath: tuple[ProvenanceNode, ...] = field(default=())
    _: KW_ONLY
    from_injection: Provenance | None = None

    # Used for memoization
    _hash: int | None = field(
        default=None, compare=False, init=False, repr=False)

    def with_appended(self, node_to_append: ProvenanceNode) -> Provenance:
        return Provenance(
            (*self.slotpath, node_to_append),
            from_injection=self.from_injection)

    def bind_content(
            self,
            name: str,
            template_preload: dict[TemplateClass, ParsedTemplateResource]
            ) -> object:
        """Use this to calculate a concrete value for use in rendering.
        This walks up the provenance stack, recursively looking for any
        overrides to the content. If none are found, it returns the
        value from the childmost instance in the provenance.
        """
        slotpath = self.slotpath
        # We use the literal ellipsis type as a sentinel for values not being
        # added yet, so we might as well just continue the trend!
        current_provenance_node = slotpath[-1]
        value = getattr(current_provenance_node.instance, name, ...)
        encloser_param_name = name
        encloser_slot_key = current_provenance_node.encloser_slot_key
        for encloser in reversed(slotpath[0:-1]):
            template_class = type(encloser.instance)
            # We do this so that env funcs that inject templates don't try
            # to continue looking up the provenance tree for slots that don't
            # actually exist.
            if hasattr(template_class, '_TEMPLATEY_EMPTY_INSTANCE'):
                break

            encloser_template = template_preload[template_class]
            encloser_overrides = (
                encloser_template.slots[encloser_slot_key].params)

            if encloser_param_name in encloser_overrides:
                value = encloser_overrides[encloser_param_name]

                if isinstance(value, TemplateInstanceContentRef):
                    encloser_slot_key = encloser.encloser_slot_key
                    encloser_param_name = value.name
                    value = ...
                else:
                    break
            else:
                break

        if value is ...:
            raise KeyError(
                'No value found for content with name at slot!',
                slotpath[-1].instance, name)

        return value

    def bind_variable(
            self,
            name: str,
            template_preload: dict[TemplateClass, ParsedTemplateResource]
            ) -> object:
        """Use this to calculate a concrete value for use in rendering.
        This walks up the provenance stack, recursively looking for any
        overrides to the variable. If none are found, it returns the
        value from the childmost instance in the provenance.
        """
        slotpath = self.slotpath
        # We use the literal ellipsis type as a sentinel for values not being
        # added yet, so we might as well just continue the trend!
        current_provenance_node = slotpath[-1]
        value = getattr(current_provenance_node.instance, name, ...)
        encloser_param_name = name
        encloser_slot_key = current_provenance_node.encloser_slot_key
        for encloser in reversed(slotpath[0:-1]):
            template_class = type(encloser.instance)
            # We do this so that env funcs that inject templates don't try
            # to continue looking up the provenance tree for slots that don't
            # actually exist.
            if hasattr(template_class, '_TEMPLATEY_EMPTY_INSTANCE'):
                break

            encloser_template = template_preload[template_class]
            encloser_overrides = (
                encloser_template.slots[encloser_slot_key].params)

            if encloser_param_name in encloser_overrides:
                value = encloser_overrides[encloser_param_name]

                if isinstance(value, TemplateInstanceVariableRef):
                    encloser_slot_key = encloser.encloser_slot_key
                    encloser_param_name = value.name
                    value = ...
                else:
                    break
            else:
                break

        if value is ...:
            raise KeyError(
                'No value found for variable with name at slot!',
                slotpath[-1].instance, name)

        return value

    @classmethod
    def from_slot_tree(
            cls,
            root_template_instance: TemplateParamsInstance,
            root_slot_tree: SlotTreeNode,
            from_injection: Provenance | None,
            ) -> list[Provenance]:
        """Given a root template instance, walks the passed slot tree,
        finding all terminus points. Returns them as a list of
        provenance instances.

        Note that you must choose the correct slot tree from the
        slot tree lookup in advance; this simply converts slot tree
        terminus points into provenances based on a root template
        instance: no more, no less.

        TODO: it would be nice if we could pre-merge the slot trees during
        template loading for the actual use cases (env functions invocations
        and, eventually, dynamic slot types) so that we didn't need a separate
        tree traversal for every template class. But for now, this is good
        enough
        """
        provenances: list[Provenance] = []
        stack: list[_TreeFlattenerFrame] = [
            _TreeFlattenerFrame(
                active_instance=root_template_instance,
                active_subtree=root_slot_tree,
                target_subtree_index=0,
                target_instance_index=0,
                wip_provenance=Provenance(
                    (
                        ProvenanceNode(
                            encloser_slot_key='',
                            encloser_slot_index=-1,
                            instance_id=id(root_template_instance),
                            instance=root_template_instance),),
                    from_injection=from_injection))]

        while stack:
            this_frame = stack[-1]
            if this_frame.exhausted:
                stack.pop()

                if this_frame.active_subtree.is_terminus:
                    provenances.append(this_frame.wip_provenance)

                continue

            this_slot_route = this_frame.active_subtree[
                this_frame.target_subtree_index]
            this_slot_name, this_slot_type, this_subtree = this_slot_route
            target_instance_index = this_frame.target_instance_index

            # This is, in a way, a nested stack, but we're maintaining
            # the stack state within the _TreeFlattenerFrame.
            # At any rate, we use the zero-index iteration of the loop
            # to memoize some values on the stack frame.
            if target_instance_index == 0:
                target_instances = getattr(
                    this_frame.active_instance,
                    this_slot_name)
                target_instances_count = len(target_instances)

                # Check in advance if there are no target instances at all,
                # and if so, skip the whole thing. This isn't just for
                # performance; the processing logic depends on it.
                if target_instances_count > 0:
                    this_frame.target_instances_count = target_instances_count
                    this_frame.target_instances = target_instances
                else:
                    # Note: this is critical! Otherwise we'll infinitely loop.
                    this_frame.target_subtree_index += 1
                    continue

            else:
                target_instances_count = this_frame.target_instances_count
                # We've exhausted the target instances; reset the state for
                # the next slot tree route and then continue.
                if this_frame.target_instance_index >= target_instances_count:
                    # Note: we're deliberately skipping the target instances
                    # themselves, because it'll just get overwritten the next
                    # time around, so we can save ourselves an operation.
                    this_frame.target_instances_count = 0
                    this_frame.target_instance_index = 0
                    # Note: this is critical! Otherwise we'll infinitely loop.
                    this_frame.target_subtree_index += 1
                    continue

                # We still have some instances to target; normalize the state
                # so that we can operate on them.
                target_instances = this_frame.target_instances

            # Okay, status check: we have our stack frame state configured
            # correctly, and we have target instances to check.
            instance_to_check = target_instances[target_instance_index]
            # Note: exact match here; not subclassing! Subclassing breaks too
            # many things, so we don't support it.
            if type(instance_to_check) is this_slot_type:
                stack.append(_TreeFlattenerFrame(
                    active_instance=instance_to_check,
                    active_subtree=this_subtree,
                    target_subtree_index=0,
                    target_instance_index=0,
                    wip_provenance=Provenance(
                        (
                            *this_frame.wip_provenance.slotpath,
                            ProvenanceNode(
                                encloser_slot_key=this_slot_name,
                                encloser_slot_index=target_instance_index,
                                instance_id=id(instance_to_check),
                                instance=instance_to_check)),
                        from_injection=from_injection)))

            this_frame.target_instance_index += 1

        return provenances

    def __hash__(self) -> int:
        memoized = self._hash
        if memoized is None:
            retval = hash(self.slotpath) ^ hash(self.from_injection)
            object.__setattr__(self, '_hash', retval)
            return retval

        else:
            return memoized


@dataclass(slots=True)
class _TreeFlattenerFrame:
    active_instance: TemplateParamsInstance
    active_subtree: SlotTreeNode
    target_subtree_index: int
    target_instance_index: int
    target_instances_count: int = field(kw_only=True, default=0)
    target_instances: Sequence[TemplateParamsInstance] = field(
        kw_only=True, init=False)
    wip_provenance: Provenance

    _active_subtree_len: int = field(init=False, repr=False, compare=False)

    def __post_init__(self):
        self._active_subtree_len = len(self.active_subtree)

    @property
    def exhausted(self) -> bool:
        return self.target_subtree_index >= self._active_subtree_len
